"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

from datetime import datetime
from typing import List, Union, Optional

import requests

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin
from platypush.schemas.maps import MapsTravelTimeSchema
from platypush.utils import to_datetime

datetime_types = Union[str, int, float, datetime]


class GoogleMapsPlugin(GooglePlugin):
    """
    Plugins that provides utilities to interact with Google Maps API services.

    Requires:

        * **google-api-python-client** (``pip install google-api-python-client``)
        * **oauth2client** (``pip install oauth2client``)

    """

    scopes = []

    def __init__(self, api_key, *args, **kwargs):
        """
        :param api_key: Server-side API key to be used for the requests, get one at
            https://console.developers.google.com
        :type api_key: str
        """

        super().__init__(scopes=self.scopes, *args, **kwargs)
        self.api_key = api_key

    @action
    def get_address_from_latlng(self, latitude, longitude):
        """
        Get an address information given lat/long

        :param latitude: Latitude
        :type latitude: float

        :param longitude: Longitude
        :type longitude: float
        """

        response = requests.get('https://maps.googleapis.com/maps/api/geocode/json',
                                params={
                                    'latlng': '{},{}'.format(latitude, longitude),
                                    'key': self.api_key,
                                }).json()

        address = dict(
            (t, None) for t in ['street_number', 'street', 'locality', 'country', 'postal_code']
        )

        address['latitude'] = latitude
        address['longitude'] = longitude

        if 'results' in response and response['results']:
            result = response['results'][0]
            self.logger.info('Google Maps geocode response for latlng ({},{}): {}'.
                             format(latitude, longitude, result))

            address['address'] = result['formatted_address'].split(',')[0]
            for addr_component in result['address_components']:
                for component_type in addr_component['types']:
                    if component_type == 'locality':
                        address['locality'] = addr_component['long_name']
                    elif component_type == 'country':
                        address['country'] = addr_component['short_name'].lower()
                    elif component_type == 'postal_code':
                        address['postal_code'] = addr_component['long_name']

        return address

    @action
    def get_elevation_from_latlng(self, latitude, longitude):
        """
        Get the elevation in meters of a geo point given lat/long

        :param latitude: Latitude
        :type latitude: float

        :param longitude: Longitude
        :type longitude: float
        """

        response = requests.get('https://maps.googleapis.com/maps/api/elevation/json',
                                params={
                                    'locations': '{},{}'.format(latitude, longitude),
                                    'key': self.api_key,
                                }).json()

        elevation = None

        if response.get('results'):
            elevation = response['results'][0]['elevation']

        return {'elevation': elevation}

    @action
    def get_travel_time(self, origins: List[str], destinations: List[str],
                        departure_time: Optional[datetime_types] = None,
                        arrival_time: Optional[datetime_types] = None,
                        units: str = 'metric',
                        avoid: Optional[List[str]] = None,
                        language: Optional[str] = None,
                        mode: Optional[str] = None,
                        traffic_model: Optional[str] = None,
                        transit_mode: Optional[List[str]] = None,
                        transit_route_preference: Optional[str] = None):
        """
        Get the estimated travel time between a set of departure points and a set of destinations.

        :param origins: The starting point(s) for calculating travel distance and time. Locations are supplied in the
            form of a place ID, an address, or latitude/longitude coordinates:

            * *Place ID*: If you supply a place ID, you must prefix it with ``place_id:``.
            * *Address*: If you pass an address, the service geocodes the string and converts it to a
              latitude/longitude coordinate to calculate distance.
            * *Coordinates*: If you pass latitude/longitude coordinates, they they will snap to the nearest road.

        :param destinations: The arrival point(s) for calculating travel distance and time. Locations are supplied in
            the form of a place ID, an address, or latitude/longitude coordinates:

            * *Place ID*: If you supply a place ID, you must prefix it with ``place_id:``.
            * *Address*: If you pass an address, the service geocodes the string and converts it to a
              latitude/longitude coordinate to calculate distance.
            * *Coordinates*: If you pass latitude/longitude coordinates, they they will snap to the nearest road.

        :param departure_time: Preferred departure time, as a UNIX timestamp, an ISO8601 string or a datetime object.
        :param arrival_time: Preferred arrival time, as a UNIX timestamp, an ISO8601 string or a datetime object.
        :param units: ``metric`` (default) or ``imperial``.
        :param avoid: Any of the following values:

            * ``tolls``: indicates that the calculated route should avoid toll roads/bridges.
            * ``highways``: indicates that the calculated route should avoid highways.
            * ``ferries``: indicates that the calculated route should avoid ferries.
            * ``indoor`` indicates that the calculated route should avoid indoor steps for walking and transit
              directions.

        :param language: Target language for the results (ISO string). If none is specified then the target language
            is inferred from the ``Accept-Language`` header or the preferred language associated to the account.
        :param mode: Any of the following:

            * ``DRIVING`` (default)
            * ``WALKING``
            * ``BICYCLING``
            * ``TRANSIT``

        :param traffic_model: Specifies the assumptions to use when calculating time in traffic. This setting affects
            the value returned in the ``duration_in_traffic`` field in the response, which contains the predicted time
            in traffic based on historical averages. Available values:

                * ``best_guess`` (default) indicates that the returned ``duration_in_traffic`` should be the best
                    estimate of travel time given what is known about both historical traffic conditions and live
                    traffic. Live traffic becomes more important the closer the departure_time is to now.
                * ``pessimistic`` indicates that the returned ``duration_in_traffic`` should be longer than the actual
                  travel time on most days, though occasional days with particularly bad traffic conditions may exceed
                  this value.
                * ``optimistic`` indicates that the returned ``duration_in_traffic`` should be shorter than the actual
                  travel time on most days, though occasional days with particularly good traffic conditions may be
                  faster than this value.

        :param transit_mode:  Specifies one or more preferred modes of transit. This parameter may only be specified
            for transit directions. Available values:

                * ``bus`` indicates that the calculated route should prefer travel by bus.
                * ``subway`` indicates that the calculated route should prefer travel by subway.
                * ``train`` indicates that the calculated route should prefer travel by train.
                * ``tram`` indicates that the calculated route should prefer travel by tram and light rail.
                * ``rail`` indicates that the calculated route should prefer travel by train, tram, light rail, and
                   subway. This is equivalent to ``transit_mode=["train", "tram", "subway"]``.

        :param transit_route_preference: Specifies preferences for transit routes. Using this parameter, you can
            bias the options returned, rather than accepting the default best route chosen by the API. This parameter
            may only be specified for transit directions. Available values:

                - ``less_walking`` indicates that the calculated route should prefer limited amounts of walking.
                - ``fewer_transfers`` indicates that the calculated route should prefer a limited number of transfers.

        :return: .. schema:: maps.MapsTravelTimeSchema
        """
        rs = requests.get(
            'https://maps.googleapis.com/maps/api/distancematrix/json',
            params={
                'origins': '|'.join(origins),
                'destinations': '|'.join(destinations),
                'units': units,
                **({'departure_time': to_datetime(departure_time)} if departure_time else {}),
                **({'arrival_time': to_datetime(arrival_time)} if arrival_time else {}),
                **({'avoid': '|'.join(avoid)} if avoid else {}),
                **({'language': language} if language else {}),
                **({'mode': mode} if mode else {}),
                **({'traffic_model': traffic_model} if traffic_model else {}),
                **({'transit_mode': transit_mode} if transit_mode else {}),
                **({'transit_route_preference': transit_route_preference}
                   if transit_route_preference else {}),
                'key': self.api_key,
            }).json()

        assert not rs.get('error_message'), f'{rs["status"]}: {rs["error_message"]}'
        rows = rs.get('rows', [])
        assert rows, 'The API returned no rows'
        elements = rows[0].get('elements')
        assert elements, 'The API returned no elements'

        return MapsTravelTimeSchema().dump(elements, many=True)


# vim:sw=4:ts=4:et:



import os
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from typing import Optional, List, Union, Dict

from platypush.plugins import Plugin, action


@dataclass
class ClientConfig:
    url: str
    username: str
    password: str

    def to_dict(self):
        return {
            'url': self.url,
            'username': self.username,
            'password': self.password,
        }


class ShareType(IntEnum):
    USER = 0
    GROUP = 1
    PUBLIC_LINK = 3
    EMAIL = 4
    FEDERATED_CLOUD_SHARE = 6
    CIRCLE = 7
    TALK_CONVERSATION = 10


class Permission(IntEnum):
    READ = 1
    UPDATE = 2
    CREATE = 4
    DELETE = 8
    SHARE = 16
    ALL = 31


class NextcloudPlugin(Plugin):
    """
    Plugin to interact with a NextCloud instance.

    Requires:

        * **nextcloud-api-wrapper** (``pip install nextcloud-api-wrapper``)

    """

    def __init__(
        self,
        url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs
    ):
        """
        :param url: URL to the index of your default NextCloud instance.
        :param username: Default NextCloud username.
        :param password: Default NextCloud password.
        """
        super().__init__(**kwargs)
        self.conf = ClientConfig(url=url, username=username, password=password)
        self._client = self._get_client(**self.conf.to_dict())

    def _get_client(
        self,
        url: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        raise_on_empty: bool = False,
    ):
        from nextcloud import NextCloud

        if not url:
            if not self.conf.url:
                if raise_on_empty:
                    raise AssertionError('No url/username/password provided')
                return None

            return NextCloud(
                endpoint=self.conf.url,
                user=self.conf.username,
                password=self.conf.password,
            )

        return NextCloud(endpoint=url, user=username, password=password)

    @staticmethod
    def _get_permissions(permissions: Optional[List[str]]) -> int:
        int_perm = 0

        for perm in permissions or []:
            perm = perm.upper()
            assert hasattr(
                Permission, perm
            ), 'Unknown permissions type: {}. Supported permissions: {}'.format(
                perm, [p.name.lower() for p in Permission]
            )

            if perm == 'ALL':
                int_perm = Permission.ALL.value
                break

            int_perm += getattr(Permission, perm).value

        return int_perm

    @staticmethod
    def _get_share_type(share_type: str) -> int:
        share_type = share_type.upper()
        assert hasattr(
            ShareType, share_type
        ), 'Unknown share type: {}. Supported share types: {}'.format(
            share_type, [s.name.lower() for s in ShareType]
        )

        return getattr(ShareType, share_type).value

    def _execute(self, server_args: dict, method: str, *args, **kwargs):
        client = self._get_client(**server_args)
        assert hasattr(client, method), 'No such NextCloud method: {}'.format(method)

        response = getattr(client, method)(*args, **kwargs)
        if response is None:
            return

        assert response.is_ok, 'Error on {method}({args}{sep}{kwargs}): {error}'.format(
            method=method,
            args=', '.join(args),
            sep=', ' if args and kwargs else '',
            kwargs=', '.join(['{}={}'.format(k, v) for k, v in kwargs.items()]),
            error=response.meta.get('message', '[No message]')
            if hasattr(response, 'meta')
            else response.raw.reason,
        )

        return response.json_data

    @action
    def get_activities(
        self,
        since: Optional[id] = None,
        limit: Optional[int] = None,
        object_type: Optional[str] = None,
        object_id: Optional[int] = None,
        sort: str = 'desc',
        **server_args
    ) -> List[str]:
        """
        Get the list of recent activities on an instance.

        :param since: Only return the activities that have occurred since the specified ID.
        :param limit: Maximum number of activities to be returned (default: ``None``).
        :param object_type: Filter by object type.
        :param object_id: Only get the activities related to a specific ``object_id``.
        :param sort: Sort mode, ``asc`` for ascending, ``desc`` for descending (default: ``desc``).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: The list of selected activities.
        """
        return self._execute(
            server_args,
            'get_activities',
            since=since,
            limit=limit,
            object_type=object_type,
            object_id=object_id,
            sort=sort,
        )

    @action
    def get_apps(self, **server_args) -> List[str]:
        """
        Get the list of apps installed on a NextCloud instance.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: The list of installed apps as strings.
        """
        return self._execute(server_args, 'get_apps').get('apps', [])

    @action
    def enable_app(self, app_id: Union[str, int], **server_args):
        """
        Enable an app.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'enable_app', app_id)

    @action
    def disable_app(self, app_id: Union[str, int], **server_args):
        """
        Disable an app.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'disable_app', app_id)

    @action
    def get_app(self, app_id: Union[str, int], **server_args) -> dict:
        """
        Provides information about an application.

        :param app_id: App ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_app', app_id)

    @action
    def get_capabilities(self, **server_args) -> dict:
        """
        Returns the capabilities of the server.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_capabilities')

    @action
    def add_group(self, group_id: Union[str], **server_args):
        """
        Create a new group.

        :param group_id: New group unique ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_group', group_id)

    @action
    def delete_group(self, group_id: Union[str], **server_args):
        """
        Delete a group.

        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_group', group_id)

    @action
    def get_group(self, group_id: Union[str], **server_args) -> dict:
        """
        Get the information of a group.

        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group', group_id)

    @action
    def get_groups(
        self,
        search: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **server_args
    ) -> List[str]:
        """
        Search for groups.

        :param search: Search for groups matching the specified substring.
        :param limit: Maximum number of returned entries.
        :param offset: Start offset.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(
            server_args, 'get_groups', search=search, limit=limit, offset=offset
        ).get('groups', [])

    @action
    def create_group_folder(self, name: str, **server_args):
        """
        Create a new group folder.

        :param name: Name/path of the folder.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'create_group_folder', name)

    @action
    def delete_group_folder(self, folder_id: Union[int, str], **server_args):
        """
        Delete a new group folder.

        :param folder_id: Folder ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_group_folder', folder_id)

    @action
    def get_group_folder(self, folder_id: Union[int, str], **server_args) -> dict:
        """
        Get a new group folder.

        :param folder_id: Folder ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group_folder', folder_id)

    @action
    def get_group_folders(self, **server_args) -> list:
        """
        Get the list new group folder.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_group_folders')

    @action
    def rename_group_folder(
        self, folder_id: Union[int, str], new_name: str, **server_args
    ):
        """
        Rename a group folder.

        :param folder_id: Folder ID.
        :param new_name: New folder name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'rename_group_folder', folder_id, new_name)

    @action
    def grant_access_to_group_folder(
        self, folder_id: Union[int, str], group_id: str, **server_args
    ):
        """
        Grant access to a group folder to a given group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'grant_access_to_group_folder', folder_id, group_id)

    @action
    def revoke_access_to_group_folder(
        self, folder_id: Union[int, str], group_id: str, **server_args
    ):
        """
        Revoke access to a group folder to a given group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'revoke_access_to_group_folder', folder_id, group_id)

    @action
    def set_group_folder_quota(
        self, folder_id: Union[int, str], quota: Optional[int], **server_args
    ):
        """
        Set the quota of a group folder.

        :param folder_id: Folder ID.
        :param quota: Quota in bytes - set None for unlimited.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(
            server_args,
            'set_quota_of_group_folder',
            folder_id,
            quota if quota is not None else -3,
        )

    @action
    def set_group_folder_permissions(
        self,
        folder_id: Union[int, str],
        group_id: str,
        permissions: List[str],
        **server_args
    ):
        """
        Set the permissions on a folder for a group.

        :param folder_id: Folder ID.
        :param group_id: Group ID.
        :param permissions: New permissions, as a list including any of the following:

          - ``read``
          - ``update``
          - ``create``
          - ``delete``
          - ``share``
          - ``all``

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(
            server_args,
            'set_permissions_to_group_folder',
            folder_id,
            group_id,
            self._get_permissions(permissions),
        )

    @action
    def get_notifications(self, **server_args) -> list:
        """
        Get the list of notifications for the logged user.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_notifications')

    @action
    def delete_notifications(self, **server_args):
        """
        Delete all notifications for the logged user.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_all_notifications')

    @action
    def get_notification(self, notification_id: int, **server_args) -> Union[dict, str]:
        """
        Get the content of a notification.

        :param notification_id: Notification ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_notification', notification_id)

    @action
    def delete_notification(self, notification_id: int, **server_args):
        """
        Delete a notification.

        :param notification_id: Notification ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_notification', notification_id)

    @action
    def create_share(
        self,
        path: str,
        share_type: str,
        share_with: Optional[str] = None,
        public_upload: bool = False,
        password: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        **server_args
    ) -> dict:
        """
        Share a file/folder with a user/group or a public link.

        :param path: Path to the resource to be shared.
        :param share_type: Share type. Supported values:

          - ``user``
          - ``group``
          - ``public_link``
          - ``email``
          - ``federated_cloud_share``
          - ``circle``
          - ``talk_conversation``

        :param share_with: User/group ID, email or conversation ID the resource should be shared with.
        :param public_upload: Whether public upload to the shared folder is allowed (default: False).
        :param password: Optional password to protect the share.
        :param permissions: Share permissions, as a list including any of the following (default: ``read``):

          - ``read``
          - ``update``
          - ``create``
          - ``delete``
          - ``share``
          - ``all``

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: The details of the newly created share. Example:

          .. code-block:: json

            {
              "id": "4",
              "share_type": 3,
              "uid_owner": "your_uid",
              "displayname_owner": "Your Name",
              "permissions": 17,
              "can_edit": true,
              "can_delete": true,
              "stime": 1599691325,
              "parent": null,
              "expiration": null,
              "token": "AbCdEfG0123456789",
              "uid_file_owner": "your_uid",
              "note": "",
              "label": "",
              "displayname_file_owner": "Your Name",
              "path": "/Shared Path",
              "item_type": "folder",
              "mimetype": "httpd/unix-directory",
              "storage_id": "home::your-uid",
              "storage": 2,
              "item_source": 13960,
              "file_source": 13960,
              "file_parent": 6,
              "file_target": "/Shared Path",
              "share_with": null,
              "share_with_displayname": "(Shared link)",
              "password": null,
              "send_password_by_talk": false,
              "url": "https://your-domain/nextcloud/index.php/s/AbCdEfG0123456789",
              "mail_send": 1,
              "hide_download": 0
            }

        """
        share_type = self._get_share_type(share_type)
        permissions = self._get_permissions(permissions or ['read'])
        return self._execute(
            server_args,
            'create_share',
            path,
            share_type=share_type,
            share_with=share_with,
            public_upload=public_upload,
            password=password,
            permissions=permissions,
        )

    @action
    def get_shares(self, **server_args) -> List[dict]:
        """
        Get the list of shares available on the server.

        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: List of available shares. Example:

          .. code-block:: json

            [
                {
                  "id": "4",
                  "share_type": 3,
                  "uid_owner": "your_uid",
                  "displayname_owner": "Your Name",
                  "permissions": 17,
                  "can_edit": true,
                  "can_delete": true,
                  "stime": 1599691325,
                  "parent": null,
                  "expiration": null,
                  "token": "AbCdEfG0123456789",
                  "uid_file_owner": "your_uid",
                  "note": "",
                  "label": "",
                  "displayname_file_owner": "Your Name",
                  "path": "/Shared Path",
                  "item_type": "folder",
                  "mimetype": "httpd/unix-directory",
                  "storage_id": "home::your-uid",
                  "storage": 2,
                  "item_source": 13960,
                  "file_source": 13960,
                  "file_parent": 6,
                  "file_target": "/Shared Path",
                  "share_with": null,
                  "share_with_displayname": "(Shared link)",
                  "password": null,
                  "send_password_by_talk": false,
                  "url": "https://your-domain/nextcloud/index.php/s/AbCdEfG0123456789",
                  "mail_send": 1,
                  "hide_download": 0
                }
            ]

        """
        return self._execute(server_args, 'get_shares')

    @action
    def delete_share(self, share_id: int, **server_args):
        """
        Remove the shared state of a resource.

        :param share_id: Share ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_share', str(share_id))

    @action
    def get_share(self, share_id: int, **server_args) -> dict:
        """
        Get the information of a shared resource.

        :param share_id: Share ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        return self._execute(server_args, 'get_share_info', str(share_id))

    @action
    def update_share(
        self,
        share_id: int,
        public_upload: Optional[bool] = None,
        password: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        expire_date: Optional[str] = None,
        **server_args
    ):
        """
        Update the permissions of a shared resource.

        :param share_id: Share ID.
        :param public_upload: Whether public upload to the shared folder is allowed (default: False).
        :param password: Optional password to protect the share.
        :param permissions: Share permissions, as a list including any of the following (default: ``read``):

          - ``read``
          - ``update``
          - ``create``
          - ``delete``
          - ``share``
          - ``all``

        :param expire_date: Share expiration date, in the format ``YYYY-MM-DD``.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        if permissions:
            permissions = self._get_permissions(permissions)

        self._execute(
            server_args,
            'update_share',
            share_id,
            public_upload=public_upload,
            password=password,
            permissions=permissions,
            expire_date=expire_date,
        )

    @action
    def create_user(self, user_id: str, password: str, **server_args):
        """
        Create a user.

        :param user_id: User ID/name.
        :param password: User password
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_user', user_id, password)

    @action
    def edit_user(self, user_id: str, properties: Dict[str, str], **server_args):
        """
        Update a set of properties of a user.

        :param user_id: User ID/name.
        :param properties: Key-value pair of user attributes to be edited.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        for k, v in properties.items():
            self._execute(server_args, 'edit_user', user_id, k, v)

    @action
    def get_user(self, user_id: str, **server_args) -> dict:
        """
        Get the details of a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: User details. Example:

          .. code-block:: json

            {
              "enabled": true,
              "storageLocation": "/mnt/hd/nextcloud/user",
              "id": "user",
              "lastLogin": 1599693750000,
              "backend": "Database",
              "subadmin": [],
              "quota": {
                "free": 6869434515456,
                "used": 1836924441,
                "total": 6871271439897,
                "relative": 0.03,
                "quota": -3
              },
              "email": "info@yourdomain.com",
              "displayname": "Your Name",
              "phone": "+1234567890",
              "address": "",
              "website": "https://yourdomain.com",
              "twitter": "@You",
              "groups": [
                "admin"
              ],
              "language": "en",
              "locale": "",
              "backendCapabilities": {
                "setDisplayName": true,
                "setPassword": true
              }
            }

        """
        return self._execute(server_args, 'get_user', user_id)

    @action
    def get_users(
        self,
        search: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **server_args
    ) -> List[str]:
        """
        Get the list of users matching some search criteria.

        :param search: Return users matching the provided string.
        :param limit: Maximum number of results to be returned (default: no limit).
        :param offset: Search results offset (default: None).
        :return: List of the matched user IDs.
        """
        return self._execute(
            server_args, 'get_users', search=search, limit=limit, offset=offset
        )

    @action
    def delete_user(self, user_id: str, **server_args):
        """
        Delete a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'delete_user', user_id)

    @action
    def enable_user(self, user_id: str, **server_args):
        """
        Enable a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'enable_user', user_id)

    @action
    def disable_user(self, user_id: str, **server_args):
        """
        Disable a user.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'disable_user', user_id)

    @action
    def add_to_group(self, user_id: str, group_id: str, **server_args):
        """
        Add a user to a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'add_to_group', user_id, group_id)

    @action
    def remove_from_group(self, user_id: str, group_id: str, **server_args):
        """
        Remove a user from a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'remove_from_group', user_id, group_id)

    @action
    def create_subadmin(self, user_id: str, group_id: str, **server_args):
        """
        Add a user as a subadmin for a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'create_subadmin', user_id, group_id)

    @action
    def remove_subadmin(self, user_id: str, group_id: str, **server_args):
        """
        Remove a user as a subadmin from a group.

        :param user_id: User ID/name.
        :param group_id: Group ID.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        self._execute(server_args, 'remove_subadmin', user_id, group_id)

    @action
    def get_subadmin_groups(self, user_id: str, **server_args) -> List[str]:
        """
        Get the groups where a given user is subadmin.

        :param user_id: User ID/name.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        :return: List of matched groups as strings.
        """
        return self._execute(server_args, 'get_subadmin_groups', user_id)

    @action
    def create_folder(self, path: str, user_id: Optional[str] = None, **server_args):
        """
        Create a folder.

        :param path: Path to the folder.
        :param user_id: User ID associated to the folder (default: same as the configured user).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        self._execute(server_args, 'create_folder', user_id, path)

    @action
    def delete_path(self, path: str, user_id: Optional[str] = None, **server_args):
        """
        Delete a file or folder.

        :param path: Path to the resource.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        self._execute(server_args, 'delete_path', user_id, path)

    @action
    def upload_file(
        self,
        remote_path: str,
        local_path: Optional[str] = None,
        content: Optional[str] = None,
        user_id: Optional[str] = None,
        timestamp: Optional[Union[datetime, int, str]] = None,
        **server_args
    ):
        """
        Upload a file.

        :param remote_path: Path to the remote resource.
        :param local_path: If set, identifies the path to the local file to be uploaded.
        :param content: If set, create a new file with this content instead of uploading an existing file.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param timestamp: File timestamp. If not specified it will be retrieved from the file info or set to ``now``
            if ``content`` is specified.
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)

        if isinstance(timestamp, str):
            timestamp = datetime.fromisoformat(timestamp)
        if isinstance(timestamp, datetime):
            timestamp = int(timestamp.timestamp())

        assert (local_path or content) and not (
            local_path and content
        ), 'Please specify either local_path or content'
        if local_path:
            method = 'upload_file'
            local_path = os.path.abspath(os.path.expanduser(local_path))
        else:
            method = 'upload_file_contents'

        return self._execute(
            server_args,
            method,
            user_id,
            local_path or content,
            remote_path,
            timestamp=timestamp,
        )

    @action
    def download_file(
        self,
        remote_path: str,
        local_path: str,
        user_id: Optional[str] = None,
        **server_args
    ):
        """
        Download a file.

        :param remote_path: Path to the remote resource.
        :param local_path: Path to the local folder.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        local_path = os.path.abspath(os.path.expanduser(local_path))
        cur_dir = os.getcwd()

        try:
            os.chdir(local_path)
            return self._execute(server_args, 'download_file', user_id, remote_path)
        finally:
            os.chdir(cur_dir)

    @action
    def list(
        self,
        path: str,
        user_id: Optional[str] = None,
        depth: int = 1,
        all_properties: bool = False,
        **server_args
    ) -> List[dict]:
        """
        List the content of a folder on the NextCloud instance.

        :param path: Remote path.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param depth: Search depth (default: 1).
        :param all_properties: Return all the file properties available (default: ``False``).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        return self._execute(
            server_args,
            'list_folders',
            user_id,
            path,
            depth=depth,
            all_properties=all_properties,
        )

    @action
    def list_favorites(
        self, path: Optional[str] = None, user_id: Optional[str] = None, **server_args
    ) -> List[dict]:
        """
        List the favorite items for a user.

        :param path: Return only the favorites under this path.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        return self._execute(server_args, 'list_folders', user_id, path)

    @action
    def mark_favorite(
        self, path: Optional[str] = None, user_id: Optional[str] = None, **server_args
    ):
        """
        Add a path to a user's favorites.

        :param path: Resource path.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        self._execute(server_args, 'set_favorites', user_id, path)

    @action
    def copy(
        self,
        path: str,
        destination: str,
        user_id: Optional[str] = None,
        overwrite: bool = False,
        **server_args
    ):
        """
        Copy a resource to another path.

        :param path: Resource path.
        :param destination: Destination path.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param overwrite: Set to ``True`` if you want to overwrite any existing file (default: ``False``).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        self._execute(
            server_args, 'copy_path', user_id, path, destination, overwrite=overwrite
        )

    @action
    def move(
        self,
        path: str,
        destination: str,
        user_id: Optional[str] = None,
        overwrite: bool = False,
        **server_args
    ):
        """
        Move a resource to another path.

        :param path: Resource path.
        :param destination: Destination path.
        :param user_id: User ID associated to the resource (default: same as the configured user).
        :param overwrite: Set to ``True`` if you want to overwrite any existing file (default: ``False``).
        :param server_args: Override the default server settings (see :meth:`._get_client` arguments).
        """
        user_id = user_id or server_args.get('username', self.conf.username)
        self._execute(
            server_args, 'move_path', user_id, path, destination, overwrite=overwrite
        )


# vim:sw=4:ts=4:et:



import os
import subprocess
import threading

from multiprocessing import Process

try:
    from websockets.exceptions import ConnectionClosed
    from websockets import serve as websocket_serve
except ImportError:
    from websockets import ConnectionClosed, serve as websocket_serve

from platypush.backend import Backend
from platypush.backend.http.app import application
from platypush.context import get_or_create_event_loop
from platypush.utils import get_ssl_server_context, set_thread_name


class HttpBackend(Backend):
    """
    The HTTP backend is a general-purpose web server.

    Example configuration:

      .. code-block:: yaml

        backend.http:
            # Default HTTP listen port
            port: 8008
            # Default websocket port
            websocket_port: 8009
            # External folders that will be exposed over `/resources/<name>`
            resource_dirs:
                photos: /mnt/hd/photos
                videos: /mnt/hd/videos
                music: /mnt/hd/music

    You can leverage this backend:

        * To execute Platypush commands via HTTP calls. In order to do so:

            * Register a user to Platypush through the web panel (usually served on ``http://host:8008/``).

            * Generate a token for your user, either through the web panel (Settings -> Generate Token) or via API:

                .. code-block:: shell

                    curl -XPOST -H 'Content-Type: application/json' -d '
                      {
                        "username": "$YOUR_USER",
                        "password": "$YOUR_PASSWORD"
                      }' http://host:8008/auth

            * Execute actions through the ``/execute`` endpoint:

                .. code-block:: shell

                    curl -XPOST -H 'Content-Type: application/json' -H "Authorization: Bearer $YOUR_TOKEN" -d '
                      {
                        "type": "request",
                        "action": "tts.say",
                        "args": {
                          "text": "This is a test"
                        }
                      }' http://host:8008/execute

        * To interact with your system (and control plugins and backends) through the Platypush web panel,
          by default available on ``http://host:8008/``. Any configured plugin that has an available panel
          plugin will be automatically added to the web panel.

        * To display a fullscreen dashboard with custom widgets.

            * Widgets are available as Vue.js components under ``platypush/backend/http/webapp/src/components/widgets``.

            * Explore their options (some may require some plugins or backends to be configured in order to work) and
              create a new dashboard template under ``~/.config/platypush/dashboards``- e.g. ``main.xml``:

                .. code-block:: xml

                    <Dashboard>
                        <!-- Display the following widgets on the same row. Each row consists of 12 columns.
                             You can specify the width of each widget either through class name (e.g. col-6 means
                             6 columns out of 12, e.g. half the size of the row) or inline style
                             (e.g. `style="width: 50%"`). -->
                        <Row>
                            <!-- Show a calendar widget with the upcoming events. It requires the `calendar` plugin to
                                 be enabled and configured. -->
                            <Calendar class="col-6" />

                            <!-- Show the current track and other playback info. It requires `music.mpd` plugin or any
                                 other music plugin enabled. -->
                            <Music class="col-3" />

                            <!-- Show current date, time and weather.
                                 It requires a `weather` plugin or backend enabled -->
                            <DateTimeWeather class="col-3" />
                        </Row>

                        <!-- Display the following widgets on a second row -->
                        <Row>
                            <!-- Show a carousel of images from a local folder. For security reasons, the folder must be
                                 explicitly exposed as an HTTP resource through the backend
                                 `resource_dirs` attribute. -->
                            <ImageCarousel class="col-6" img-dir="/mnt/hd/photos/carousel" />

                            <!-- Show the news headlines parsed from a list of RSS feed and stored locally through the
                                 `http.poll` backend -->
                            <RssNews class="col-6" db="sqlite:////path/to/your/rss.db" />
                        </Row>
                    </Dashboard>

            * The dashboard will be accessible under ``http://host:8008/dashboard/<name>``, where ``name=main`` if for
              example you stored your template under ``~/.config/platypush/dashboards/main.xml``.

        * To expose custom endpoints that can be called as web hooks by other applications and run some custom logic.
          All you have to do in this case is to create a hook on a
          :class:`platypush.message.event.http.hook.WebhookEvent` with the endpoint that you want to expose and store
          it under e.g. ``~/.config/platypush/scripts/hooks.py``:

            .. code-block:: python

                from platypush.context import get_plugin
                from platypush.event.hook import hook
                from platypush.message.event.http.hook import WebhookEvent

                hook_token = 'abcdefabcdef'

                # Expose the hook under the /hook/lights_toggle endpoint
                @hook(WebhookEvent, hook='lights_toggle')
                def lights_toggle(event, **context):
                    # Do any checks on the request
                    assert event.headers.get('X-Token') == hook_token, 'Unauthorized'

                    # Run some actions
                    lights = get_plugin('light.hue')
                    lights.toggle()

    Any plugin can register custom routes under ``platypush/backend/http/app/routes/plugins``.
    Any additional route is managed as a Flask blueprint template and the `.py`
    module can expose lists of routes to the main webapp through the
    ``__routes__`` object (a list of Flask blueprints).

    Security: Access to the endpoints requires at least one user to be registered. Access to the endpoints is regulated
    in the following ways (with the exception of event hooks, whose logic is up to the user):

        * **Simple authentication** - i.e. registered username and password.
        * **JWT token** provided either over as ``Authorization: Bearer`` header or ``GET`` ``?token=<TOKEN>``
          parameter. A JWT token can be generated either through the web panel or over the ``/auth`` endpoint.
        * **Global platform token**, usually configured on the root of the ``config.yaml`` as ``token: <VALUE>``.
          It can provided either over on the ``X-Token`` header or as a ``GET`` ``?token=<TOKEN>`` parameter.
        * **Session token**, generated upon login, it can be used to authenticate requests through the ``Cookie`` header
          (cookie name: ``session_token``).

    Requires:

        * **gunicorn** (``pip install gunicorn``) - optional, to run the Platypush webapp over uWSGI.

    By default the Platypush web server will run in a
    process spawned on the fly by the HTTP backend. However, being a
    Flask app, it will serve clients in a single thread and it won't
    support many features of a full-blown web server. gunicorn allows
    you to easily spawn the web server in a uWSGI wrapper, separate
    from the main Platypush daemon, and the uWSGI layer can be easily
    exposed over an nginx/lighttpd web server.

    Command to run the web server over a gunicorn uWSGI wrapper::

        gunicorn -w <n_workers> -b <bind_address>:8008 platypush.backend.http.uwsgi

    """

    _DEFAULT_HTTP_PORT = 8008
    _DEFAULT_WEBSOCKET_PORT = 8009

    def __init__(
        self,
        port=_DEFAULT_HTTP_PORT,
        websocket_port=_DEFAULT_WEBSOCKET_PORT,
        bind_address='0.0.0.0',
        disable_websocket=False,
        resource_dirs=None,
        ssl_cert=None,
        ssl_key=None,
        ssl_cafile=None,
        ssl_capath=None,
        maps=None,
        run_externally=False,
        uwsgi_args=None,
        **kwargs
    ):
        """
        :param port: Listen port for the web server (default: 8008)
        :type port: int

        :param websocket_port: Listen port for the websocket server (default: 8009)
        :type websocket_port: int

        :param bind_address: Address/interface to bind to (default: 0.0.0.0, accept connection from any IP)
        :type bind_address: str

        :param disable_websocket: Disable the websocket interface (default: False)
        :type disable_websocket: bool

        :param ssl_cert: Set it to the path of your certificate file if you want to enable HTTPS (default: None)
        :type ssl_cert: str

        :param ssl_key: Set it to the path of your key file if you want to enable HTTPS (default: None)
        :type ssl_key: str

        :param ssl_cafile: Set it to the path of your certificate authority file if you want to enable HTTPS
            (default: None)
        :type ssl_cafile: str

        :param ssl_capath: Set it to the path of your certificate authority directory if you want to enable HTTPS
            (default: None)
        :type ssl_capath: str

        :param resource_dirs: Static resources directories that will be
            accessible through ``/resources/<path>``. It is expressed as a map
            where the key is the relative path under ``/resources`` to expose and
            the value is the absolute path to expose.
        :type resource_dirs: dict[str, str]

        :param run_externally: If set, then the HTTP backend will not directly
            spawn the web server. Set this option if you plan to run the webapp
            in a separate web server (recommended), like uwsgi or uwsgi+nginx.
        :type run_externally: bool

        :param uwsgi_args: If ``run_externally`` is set and you would like the
            HTTP backend to directly spawn and control the uWSGI application
            server instance, then pass the list of uWSGI arguments through
            this parameter. Some examples include::

                # Start uWSGI instance listening on HTTP port 8008 with 4
                # processes
                ['--plugin', 'python', '--http-socket', ':8008', '--master', '--processes', '4']

                # Start uWSGI instance listening on uWSGI socket on port 3031.
                # You can then use another full-blown web server, like nginx
                # or Apache, to communicate with the uWSGI instance
                ['--plugin', 'python', '--socket', '127.0.0.1:3031', '--master', '--processes', '4']
        :type uwsgi_args: list[str]
        """

        super().__init__(**kwargs)

        self.port = port
        self.websocket_port = websocket_port
        self.maps = maps or {}
        self.server_proc = None
        self.disable_websocket = disable_websocket
        self.websocket_thread = None
        self._websocket_loop = None
        self.bind_address = bind_address

        if resource_dirs:
            self.resource_dirs = {
                name: os.path.abspath(os.path.expanduser(d))
                for name, d in resource_dirs.items()
            }
        else:
            self.resource_dirs = {}

        self.active_websockets = set()
        self.run_externally = run_externally
        self.uwsgi_args = uwsgi_args or []
        self.ssl_context = (
            get_ssl_server_context(
                ssl_cert=ssl_cert,
                ssl_key=ssl_key,
                ssl_cafile=ssl_cafile,
                ssl_capath=ssl_capath,
            )
            if ssl_cert
            else None
        )

        if self.uwsgi_args:
            self.uwsgi_args = [str(_) for _ in self.uwsgi_args] + [
                '--module',
                'platypush.backend.http.uwsgi',
                '--enable-threads',
            ]

        self.local_base_url = '{proto}://localhost:{port}'.format(
            proto=('https' if ssl_cert else 'http'), port=self.port
        )

        self._websocket_lock_timeout = 10
        self._websocket_lock = threading.RLock()
        self._websocket_locks = {}

    def send_message(self, msg, **kwargs):
        self.logger.warning('Use cURL or any HTTP client to query the HTTP backend')

    def on_stop(self):
        """On backend stop"""
        super().on_stop()
        self.logger.info('Received STOP event on HttpBackend')

        if self.server_proc:
            if isinstance(self.server_proc, subprocess.Popen):
                self.server_proc.kill()
                self.server_proc.wait(timeout=10)
                if self.server_proc.poll() is not None:
                    self.logger.info(
                        'HTTP server process may be still alive at termination'
                    )
                else:
                    self.logger.info('HTTP server process terminated')
            else:
                self.server_proc.terminate()
                self.server_proc.join(timeout=10)
                if self.server_proc.is_alive():
                    self.server_proc.kill()
                if self.server_proc.is_alive():
                    self.logger.info(
                        'HTTP server process may be still alive at termination'
                    )
                else:
                    self.logger.info('HTTP server process terminated')

        if (
            self.websocket_thread
            and self.websocket_thread.is_alive()
            and self._websocket_loop
        ):
            self._websocket_loop.stop()
            self.logger.info('HTTP websocket service terminated')

    def _acquire_websocket_lock(self, ws):
        try:
            acquire_ok = self._websocket_lock.acquire(
                timeout=self._websocket_lock_timeout
            )
            if not acquire_ok:
                raise TimeoutError('Websocket lock acquire timeout')

            addr = ws.remote_address
            if addr not in self._websocket_locks:
                self._websocket_locks[addr] = threading.RLock()
        finally:
            self._websocket_lock.release()

        acquire_ok = self._websocket_locks[addr].acquire(
            timeout=self._websocket_lock_timeout
        )
        if not acquire_ok:
            raise TimeoutError(
                'Websocket on address {} not ready to receive data'.format(addr)
            )

    def _release_websocket_lock(self, ws):
        try:
            acquire_ok = self._websocket_lock.acquire(
                timeout=self._websocket_lock_timeout
            )
            if not acquire_ok:
                raise TimeoutError('Websocket lock acquire timeout')

            addr = ws.remote_address
            if addr in self._websocket_locks:
                self._websocket_locks[addr].release()
        except Exception as e:
            self.logger.warning(
                'Unhandled exception while releasing websocket lock: {}'.format(str(e))
            )
        finally:
            self._websocket_lock.release()

    def notify_web_clients(self, event):
        """Notify all the connected web clients (over websocket) of a new event"""

        async def send_event(ws):
            try:
                self._acquire_websocket_lock(ws)
                await ws.send(str(event))
            except Exception as e:
                self.logger.warning('Error on websocket send_event: {}'.format(e))
            finally:
                self._release_websocket_lock(ws)

        loop = get_or_create_event_loop()
        wss = self.active_websockets.copy()

        for _ws in wss:
            try:
                loop.run_until_complete(send_event(_ws))
            except ConnectionClosed:
                self.logger.warning(
                    'Websocket client {} connection lost'.format(_ws.remote_address)
                )
                self.active_websockets.remove(_ws)
                if _ws.remote_address in self._websocket_locks:
                    del self._websocket_locks[_ws.remote_address]

    def websocket(self):
        """Websocket main server"""
        set_thread_name('WebsocketServer')

        async def register_websocket(websocket, path):
            address = (
                websocket.remote_address
                if websocket.remote_address
                else '<unknown client>'
            )

            self.logger.info(
                'New websocket connection from {} on path {}'.format(address, path)
            )
            self.active_websockets.add(websocket)

            try:
                await websocket.recv()
            except ConnectionClosed:
                self.logger.info(
                    'Websocket client {} closed connection'.format(address)
                )
                self.active_websockets.remove(websocket)
                if address in self._websocket_locks:
                    del self._websocket_locks[address]

        websocket_args = {}
        if self.ssl_context:
            websocket_args['ssl'] = self.ssl_context

        self._websocket_loop = get_or_create_event_loop()
        self._websocket_loop.run_until_complete(
            websocket_serve(
                register_websocket,
                self.bind_address,
                self.websocket_port,
                **websocket_args
            )
        )
        self._websocket_loop.run_forever()

    def _start_web_server(self):
        def proc():
            self.logger.info('Starting local web server on port {}'.format(self.port))
            kwargs = {
                'host': self.bind_address,
                'port': self.port,
                'use_reloader': False,
                'debug': False,
            }

            application.config['redis_queue'] = self.bus.redis_queue
            if self.ssl_context:
                kwargs['ssl_context'] = self.ssl_context

            application.run(**kwargs)

        return proc

    def run(self):
        super().run()
        try:
            self.register_service(port=self.port)
        except Exception as e:
            self.logger.warning('Could not register the Zeroconf service')
            self.logger.exception(e)

        if not self.disable_websocket:
            self.logger.info('Initializing websocket interface')
            self.websocket_thread = threading.Thread(target=self.websocket)
            self.websocket_thread.start()

        if not self.run_externally:
            self.server_proc = Process(
                target=self._start_web_server(), name='WebServer'
            )
            self.server_proc.start()
            self.server_proc.join()
        elif self.uwsgi_args:
            uwsgi_cmd = ['uwsgi'] + self.uwsgi_args
            self.logger.info('Starting uWSGI with arguments {}'.format(uwsgi_cmd))
            self.server_proc = subprocess.Popen(uwsgi_cmd)
        else:
            self.logger.info(
                'The web server is configured to be launched externally but '
                + 'no uwsgi_args were provided. Make sure that you run another external service'
                + 'for the webserver (e.g. nginx)'
            )


# vim:sw=4:ts=4:et:



from typing import Optional, Union, List, Dict, Any

import requests

from platypush.plugins import Plugin, action


class Covid19Plugin(Plugin):
    """
    Monitor the diffusion data of the COVID-19 pandemic by using the public API at https://api.covid19api.com.
    """

    base_url = 'https://api.covid19api.com'

    def __init__(self, country: Union[str, List[str]] = 'world', **kwargs):
        """
        :param country: Default country (or list of countries) to retrieve the stats for. It can either be the full
            country name or the country code. Special values:

            - ``world``: Get worldwide stats (default).
            - ``all``: Get all the available stats.
        """
        super().__init__(**kwargs)
        self.country = []
        self.all_countries = requests.get('{}/countries'.format(self.base_url)).json()
        self.country = self._get_countries(country)

    def _get_countries(self, country: Optional[Union[str, List[str]]] = None) -> List[str]:
        country = country or self.country
        if isinstance(country, str):
            country = country.split(',')
        lc_country = {c.lower() for c in country}
        return [c['ISO2'] for c in self.all_countries
                if c['ISO2'].lower() in lc_country
                or c['Slug'].lower() in lc_country
                or c['Country'].lower() in lc_country]

    @action
    def summary(self, country: Optional[Union[str, List[str]]] = None) -> List[Dict[str, Any]]:
        """
        Get the summary data for the world or a country.

        :param country: Default country override.
        """
        countries = self._get_countries(country)
        response = requests.get('{}/summary'.format(self.base_url)).json()
        if countries[0] == 'all':
            return response.get('Countries', [])
        if countries[0] == 'world':
            return response.get('Global', {})

        return [
            c for c in response.get('Countries', [])
            if c.get('CountryCode').upper() in countries
            or c.get('Country').upper() in countries
        ]

    @action
    def data(self, country: Optional[Union[str, List[str]]] = None) -> List[Dict[str, Any]]:
        """
        Get all the data for a country.

        :param country: Default country override.
        """
        countries = self._get_countries(country)
        ret = []
        for country in countries:
            ret += requests.get('{}/total/country/{}'.format(self.base_url, country)).json()
        return ret


# vim:sw=4:ts=4:et:



import logging
import threading

from redis import Redis

from platypush.bus import Bus
from platypush.config import Config
from platypush.message import Message

logger = logging.getLogger('platypush:bus:redis')


class RedisBus(Bus):
    """ Overrides the in-process in-memory local bus with a Redis bus """
    _DEFAULT_REDIS_QUEUE = 'platypush/bus'

    def __init__(self, *args, on_message=None, redis_queue=None, **kwargs):
        super().__init__(on_message=on_message)

        if not args and not kwargs:
            kwargs = (Config.get('backend.redis') or {}).get('redis_args', {})

        self.redis = Redis(*args, **kwargs)
        self.redis_args = kwargs
        self.redis_queue = redis_queue or self._DEFAULT_REDIS_QUEUE
        self.on_message = on_message
        self.thread_id = threading.get_ident()

    def get(self):
        """ Reads one message from the Redis queue """
        try:
            if self.should_stop():
                return

            msg = self.redis.blpop(self.redis_queue, timeout=1)
            if not msg or msg[1] is None:
                return

            msg = msg[1].decode('utf-8')
            return Message.build(msg)
        except Exception as e:
            logger.exception(e)

    def post(self, msg):
        """ Sends a message to the Redis queue """
        return self.redis.rpush(self.redis_queue, str(msg))

    def stop(self):
        super().stop()
        self.redis.close()


# vim:sw=4:ts=4:et:



import datetime
import queue
import pytest
import time

from dateutil.tz import gettz
from mock import patch

test_timeout = 10
cron_queue = queue.Queue()


class MockDatetime(datetime.datetime):
    timedelta = datetime.timedelta()

    @classmethod
    def now(cls):
        return super().now(tz=gettz()) + cls.timedelta


def _test_cron_queue(expected_msg: str):
    msg = None
    test_start = time.time()
    while time.time() - test_start <= test_timeout and msg != expected_msg:
        try:
            msg = cron_queue.get(block=True, timeout=test_timeout)
        except queue.Empty:
            break

    assert msg == expected_msg, 'The expected cronjob has not been executed'


def test_cron_execution():
    """
    Test that the cronjob in ``../etc/scripts/test_cron.py`` runs successfully.
    """
    _test_cron_queue('cron_test')


def test_cron_execution_upon_system_clock_change():
    """
    Test that the cronjob runs at the right time even upon DST or other
    system clock changes.
    """
    # Mock datetime.datetime with a class that has overridable timedelta
    patcher = patch('datetime.datetime', MockDatetime)

    try:
        patcher.start()
        time.sleep(1)
        # Simulate a +1hr shift on the system clock
        MockDatetime.timedelta = datetime.timedelta(hours=1)
        time.sleep(1)
    finally:
        patcher.stop()

    # Ensure that the cronjob that was supposed to run in an hour is now running
    _test_cron_queue('cron_1hr_test')


if __name__ == '__main__':
    pytest.main()


# vim:sw=4:ts=4:et:



from abc import ABC
from base64 import b64encode
from typing import Optional

from platypush.message.event import Event


class IRCEvent(Event, ABC):
    """
    IRC base event.
    """
    def __init__(self, *args, server: Optional[str] = None, port: Optional[int] = None,
                 alias: Optional[str] = None, channel: Optional[str] = None, **kwargs):
        super().__init__(*args, server=server, port=port, alias=alias, channel=channel, **kwargs)


class IRCChannelJoinEvent(IRCEvent):
    """
    Event triggered upon account channel join.
    """
    def __init__(self, *args, nick: str, **kwargs):
        super().__init__(*args, nick=nick, **kwargs)


class IRCChannelKickEvent(IRCEvent):
    """
    Event triggered upon account channel kick.
    """
    def __init__(self, *args, target_nick: str, source_nick: Optional[str] = None, **kwargs):
        super().__init__(*args, source_nick=source_nick, target_nick=target_nick, **kwargs)


class IRCModeEvent(IRCEvent):
    """
    Event triggered when the IRC mode of a channel user changes.
    """
    def __init__(
            self, *args, mode: str, channel: Optional[str] = None,
            source: Optional[str] = None,
            target_: Optional[str] = None, **kwargs
    ):
        super().__init__(*args, mode=mode, channel=channel, source=source, target_=target_, **kwargs)


class IRCPartEvent(IRCEvent):
    """
    Event triggered when an IRC nick parts.
    """
    def __init__(self, *args, nick: str, **kwargs):
        super().__init__(*args, nick=nick, **kwargs)


class IRCQuitEvent(IRCEvent):
    """
    Event triggered when an IRC nick quits.
    """
    def __init__(self, *args, nick: str, **kwargs):
        super().__init__(*args, nick=nick, **kwargs)


class IRCNickChangeEvent(IRCEvent):
    """
    Event triggered when a IRC nick changes.
    """
    def __init__(self, *args, before: str, after: str, **kwargs):
        super().__init__(*args, before=before, after=after, **kwargs)


class IRCConnectEvent(IRCEvent):
    """
    Event triggered upon server connection.
    """


class IRCDisconnectEvent(IRCEvent):
    """
    Event triggered upon server disconnection.
    """


class IRCPrivateMessageEvent(IRCEvent):
    """
    Event triggered when a private message is received.
    """
    def __init__(self, *args, text: str, nick: str, mentions_me: bool = False, **kwargs):
        super().__init__(*args, text=text, nick=nick, mentions_me=mentions_me, **kwargs)


class IRCPublicMessageEvent(IRCEvent):
    """
    Event triggered when a public message is received.
    """
    def __init__(self, *args, text: str, nick: str, mentions_me: bool = False, **kwargs):
        super().__init__(*args, text=text, nick=nick, mentions_me=mentions_me, **kwargs)


class IRCDCCRequestEvent(IRCEvent):
    """
    Event triggered when a DCC connection request is received.
    """
    def __init__(self, *args, address: str, port: int, nick: str, **kwargs):
        super().__init__(*args, address=address, port=port, nick=nick, **kwargs)


class IRCDCCMessageEvent(IRCEvent):
    """
    Event triggered when a DCC message is received.
    """
    def __init__(self, *args, address: str, body: bytes, **kwargs):
        super().__init__(
            *args, address=address, body=b64encode(body).decode(), **kwargs
        )


class IRCCTCPMessageEvent(IRCEvent):
    """
    Event triggered when a CTCP message is received.
    """
    def __init__(self, *args, address: str, message: str, **kwargs):
        super().__init__(*args, address=address, message=message, **kwargs)


class IRCDCCFileRequestEvent(IRCEvent):
    """
    Event triggered when a DCC file send request is received.
    """
    def __init__(
            self, *args, nick: str, address: str, file: str,
            port: int, size: Optional[int] = None, **kwargs
    ):
        super().__init__(
            *args, nick=nick, address=address, file=file, port=port,
            size=size, **kwargs
        )


class IRCDCCFileRecvCompletedEvent(IRCEvent):
    """
    Event triggered when a DCC file transfer RECV is completed.
    """
    def __init__(
            self, *args, address: str, port: int, file: str,
            size: Optional[int] = None, **kwargs
    ):
        super().__init__(
            *args, address=address, file=file,
            port=port, size=size, **kwargs
        )


class IRCDCCFileRecvCancelledEvent(IRCEvent):
    """
    Event triggered when a DCC file transfer RECV is cancelled.
    """
    def __init__(
            self, *args, address: str, port: int, file: str,
            error: str, **kwargs
    ):
        super().__init__(
            *args, address=address, file=file, port=port,
            error=error, **kwargs
        )


class IRCDCCFileSendCompletedEvent(IRCEvent):
    """
    Event triggered when a DCC file transfer SEND is completed.
    """
    def __init__(self, *args, address: str, port: int, file: str, **kwargs):
        super().__init__(*args, address=address, file=file, port=port, **kwargs)


class IRCDCCFileSendCancelledEvent(IRCEvent):
    """
    Event triggered when a DCC file transfer SEND is cancelled.
    """
    def __init__(
            self, *args, address: str, port: int, file: str,
            error: str, **kwargs
    ):
        super().__init__(
            *args, address=address, file=file, port=port,
            error=error, **kwargs
        )



from platypush.context import get_plugin
from platypush.plugins.media.search import MediaSearcher


class PlexMediaSearcher(MediaSearcher):
    def search(self, query, **kwargs):
        """
        Performs a Plex search using the configured :class:`platypush.plugins.media.plex.MediaPlexPlugin` instance if
        it is available.
        """

        try:
            plex = get_plugin('media.plex')
        except RuntimeError:
            return []

        self.logger.info('Searching Plex for "{}"'.format(query))
        results = []

        for result in plex.search(title=query).output:
            results.extend(self._flatten_result(result))

        self.logger.info('{} Plex results found for the search query "{}"'.format(len(results), query))
        return results

    @staticmethod
    def _flatten_result(result):
        def parse_part(media, part, episode=None, sub_media=None):
            if 'episodes' in media:
                del media['episodes']

            return {
                **{k: v for k, v in result.items() if k not in ['media', 'type']},
                'media_type': result.get('type'),
                'type': 'plex',
                **{k: v for k, v in media.items() if k not in ['parts']},
                **part,
                'title': '{}{}{}'.format(
                    result.get('title', ''),
                    ' [{}]'.format(episode['season_episode']) if (episode or {}).get('season_episode') else '',
                    ' {}'.format(sub_media['title']) if (sub_media or {}).get('title') else '',
                ),
                'summary': episode['summary'] if (episode or {}).get('summary') else media.get('summary'),
            }

        results = []

        for media in result.get('media', []):
            if 'episodes' in media:
                for episode in media['episodes']:
                    for sub_media in episode.get('media', []):
                        for part in sub_media.get('parts', []):
                            results.append(parse_part(media=media, episode=episode, sub_media=sub_media, part=part))
            else:
                for part in media.get('parts', []):
                    results.append(parse_part(media=media, part=part))

        return results


# vim:sw=4:ts=4:et:



import json

from flask import Blueprint, abort, request, make_response

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import logger, send_message
from platypush.config import Config
from platypush.event.hook import EventCondition
from platypush.message.event.http.hook import WebhookEvent


hook = Blueprint('hook', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    hook,
]


def matches_condition(event: WebhookEvent, hook):
    if isinstance(hook, dict):
        if_ = hook['if'].copy()
        if_['type'] = '.'.join([event.__module__, event.__class__.__qualname__])

        condition = EventCondition.build(if_)
    else:
        condition = hook.condition

    return event.matches_condition(condition)


@hook.route(
    '/hook/<hook_name>', methods=['GET', 'POST', 'PUT', 'DELETE', 'PATCH', 'OPTIONS']
)
def hook_route(hook_name):
    """Endpoint for custom webhooks"""

    event_args = {
        'hook': hook_name,
        'method': request.method,
        'args': dict(request.args or {}),
        'data': request.data.decode(),
        'headers': dict(request.headers or {}),
    }

    if event_args['data']:
        try:
            event_args['data'] = json.loads(event_args['data'])
        except Exception as e:
            logger().warning(
                'Not a valid JSON string: %s: %s', event_args['data'], str(e)
            )

    event = WebhookEvent(**event_args)
    matching_hooks = [
        hook
        for hook in Config.get_event_hooks().values()
        if matches_condition(event, hook)
    ]

    try:
        send_message(event)
        rs = default_rs = make_response(json.dumps({'status': 'ok', **event_args}))
        headers = {}
        status_code = 200

        # If there are matching hooks, wait for their completion before returning
        if matching_hooks:
            rs = event.wait_response(timeout=60)
            try:
                rs = json.loads(rs.decode())  # type: ignore
            except Exception:
                pass

            if isinstance(rs, dict) and '___data___' in rs:
                # data + http_code + custom_headers return format
                headers = rs.get('___headers___', {})
                status_code = rs.get('___code___', status_code)
                rs = rs['___data___']

            if rs is None:
                rs = default_rs
                headers = {'Content-Type': 'application/json'}

            rs = make_response(rs)
        else:
            headers = {'Content-Type': 'application/json'}

        rs.status_code = status_code
        rs.headers.update(headers)
        return rs
    except Exception as e:
        logger().exception(e)
        logger().error('Error while dispatching webhook event %s: %s', event, str(e))
        abort(500, str(e))


# vim:sw=4:ts=4:et:



import base64
import datetime
import hashlib
import json
import random
import rsa
import time
from typing import Optional, Dict

import bcrypt

from sqlalchemy import Column, Integer, String, DateTime, ForeignKey
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base

from platypush.context import get_plugin
from platypush.exceptions.user import InvalidJWTTokenException, InvalidCredentialsException
from platypush.utils import get_or_generate_jwt_rsa_key_pair

Base = declarative_base()


class UserManager:
    """
    Main class for managing platform users
    """

    # noinspection PyProtectedMember
    def __init__(self):
        db_plugin = get_plugin('db')
        if not db_plugin:
            raise ModuleNotFoundError('Please enable/configure the db plugin for multi-user support')

        self._engine = db_plugin._get_engine()

    def get_user(self, username):
        session = self._get_db_session()
        user = self._get_user(session, username)
        if not user:
            return None

        # Hide password
        user.password = None
        return user

    def get_user_count(self):
        session = self._get_db_session()
        return session.query(User).count()

    def get_users(self):
        session = self._get_db_session()
        return session.query(User)

    def create_user(self, username, password, **kwargs):
        session = self._get_db_session()
        if not username:
            raise ValueError('Invalid or empty username')
        if not password:
            raise ValueError('Please provide a password for the user')

        user = self._get_user(session, username)
        if user:
            raise NameError('The user {} already exists'.format(username))

        record = User(username=username, password=self._encrypt_password(password),
                      created_at=datetime.datetime.utcnow(), **kwargs)

        session.add(record)
        session.commit()
        user = self._get_user(session, username)

        # Hide password
        user.password = None
        return user

    def update_password(self, username, old_password, new_password):
        session = self._get_db_session()
        if not self._authenticate_user(session, username, old_password):
            return False

        user = self._get_user(session, username)
        user.password = self._encrypt_password(new_password)
        session.commit()
        return True

    def authenticate_user(self, username, password):
        session = self._get_db_session()
        return self._authenticate_user(session, username, password)

    def authenticate_user_session(self, session_token):
        session = self._get_db_session()
        user_session = session.query(UserSession).filter_by(session_token=session_token).first()

        if not user_session or (
                user_session.expires_at and user_session.expires_at < datetime.datetime.utcnow()):
            return None, None

        user = session.query(User).filter_by(user_id=user_session.user_id).first()

        # Hide password
        user.password = None
        return user, session

    def delete_user(self, username):
        session = self._get_db_session()
        user = self._get_user(session, username)
        if not user:
            raise NameError('No such user: {}'.format(username))

        user_sessions = session.query(UserSession).filter_by(user_id=user.user_id).all()
        for user_session in user_sessions:
            session.delete(user_session)

        session.delete(user)
        session.commit()
        return True

    def delete_user_session(self, session_token):
        session = self._get_db_session()
        user_session = session.query(UserSession).filter_by(session_token=session_token).first()

        if not user_session:
            return False

        session.delete(user_session)
        session.commit()
        return True

    def create_user_session(self, username, password, expires_at=None):
        session = self._get_db_session()
        user = self._authenticate_user(session, username, password)
        if not user:
            return None

        if expires_at:
            if isinstance(expires_at, int) or isinstance(expires_at, float):
                expires_at = datetime.datetime.fromtimestamp(expires_at)
            elif isinstance(expires_at, str):
                expires_at = datetime.datetime.fromisoformat(expires_at)

        user_session = UserSession(user_id=user.user_id, session_token=self.generate_session_token(),
                                   csrf_token=self.generate_session_token(), created_at=datetime.datetime.utcnow(),
                                   expires_at=expires_at)

        session.add(user_session)
        session.commit()
        return user_session

    @staticmethod
    def _get_user(session, username):
        return session.query(User).filter_by(username=username).first()

    @staticmethod
    def _encrypt_password(pwd):
        if isinstance(pwd, str):
            pwd = pwd.encode()
        return bcrypt.hashpw(pwd, bcrypt.gensalt(12)).decode()

    @classmethod
    def _check_password(cls, pwd, hashed_pwd):
        return bcrypt.checkpw(cls._to_bytes(pwd), cls._to_bytes(hashed_pwd))

    @staticmethod
    def _to_bytes(data) -> bytes:
        if isinstance(data, str):
            data = data.encode()
        return data

    @staticmethod
    def generate_session_token():
        rand = bytes(random.randint(0, 255) for _ in range(0, 255))
        return hashlib.sha256(rand).hexdigest()

    def get_user_by_session(self, session_token: str):
        """
        Get a user associated to a session token.

        :param session_token: Session token.
        """
        session = self._get_db_session()
        return session.query(User).join(UserSession).filter_by(session_token=session_token).first()

    def generate_jwt_token(self, username: str, password: str, expires_at: Optional[datetime.datetime] = None) -> str:
        """
        Create a user JWT token for API usage.

        :param username: User name.
        :param password: Password.
        :param expires_at: Expiration datetime of the token.
        :return: The generated JWT token as a string.
        :raises: :class:`platypush.exceptions.user.InvalidCredentialsException` in case of invalid credentials.
        """
        user = self.authenticate_user(username, password)
        if not user:
            raise InvalidCredentialsException()

        pub_key, _ = get_or_generate_jwt_rsa_key_pair()
        payload = json.dumps(
            {
                'username': username,
                'password': password,
                'created_at': datetime.datetime.now().timestamp(),
                'expires_at': expires_at.timestamp() if expires_at else None,
            },
            sort_keys=True,
            indent=None,
        )

        return base64.b64encode(
            rsa.encrypt(payload.encode('ascii'), pub_key)
        ).decode()

    def validate_jwt_token(self, token: str) -> Dict[str, str]:
        """
        Validate a JWT token.

        :param token: Token to validate.
        :return: On success, it returns the JWT payload with the following structure:

            .. code-block:: json

                {
                    "username": "user ID/name",
                    "created_at": "token creation timestamp",
                    "expires_at": "token expiration timestamp"
                }

        :raises: :class:`platypush.exceptions.user.InvalidJWTTokenException` in case of invalid token.
        """
        _, priv_key = get_or_generate_jwt_rsa_key_pair()

        try:
            payload = json.loads(
                rsa.decrypt(
                    base64.b64decode(token.encode('ascii')),
                    priv_key
                ).decode('ascii')
            )
        except (TypeError, ValueError) as e:
            raise InvalidJWTTokenException(f'Could not decode JWT token: {e}')

        expires_at = payload.get('expires_at')
        if expires_at and time.time() > expires_at:
            raise InvalidJWTTokenException('Expired JWT token')

        user = self.authenticate_user(
            payload.get('username', ''),
            payload.get('password', '')
        )

        if not user:
            raise InvalidCredentialsException()

        return payload

    def _get_db_session(self):
        Base.metadata.create_all(self._engine)
        session = scoped_session(sessionmaker(expire_on_commit=False))
        session.configure(bind=self._engine)
        return session()

    def _authenticate_user(self, session, username, password):
        """
        :return: :class:`platypush.user.User` instance if the user exists and the password is valid, ``None`` otherwise.
        """
        user = self._get_user(session, username)
        if not user:
            return None

        if not self._check_password(password, user.password):
            return None

        return user


class User(Base):
    """ Models the User table """

    __tablename__ = 'user'
    __table_args__ = ({'sqlite_autoincrement': True})

    user_id = Column(Integer, primary_key=True)
    username = Column(String, unique=True, nullable=False)
    password = Column(String)
    created_at = Column(DateTime)


class UserSession(Base):
    """ Models the UserSession table """

    __tablename__ = 'user_session'
    __table_args__ = ({'sqlite_autoincrement': True})

    session_id = Column(Integer, primary_key=True)
    session_token = Column(String, unique=True, nullable=False)
    csrf_token = Column(String, unique=True)
    user_id = Column(Integer, ForeignKey('user.user_id'), nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)


# vim:sw=4:ts=4:et:



import requests
from urllib.parse import urljoin

from flask import abort, request, Blueprint

from platypush.backend.http.app import template_folder

mimic3 = Blueprint('mimic3', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    mimic3,
]


@mimic3.route('/tts/mimic3/say', methods=['GET'])
def proxy_tts_request():
    """
    This route is used to proxy the POST request to the Mimic3 TTS server
    through a GET, so it can be easily processed as a URL through a media
    plugin.
    """
    required_args = {
        'text',
        'server_url',
        'voice',
    }

    missing_args = required_args.difference(set(request.args.keys()))
    if missing_args:
        abort(400, f'Missing parameters: {missing_args}')

    args = {arg: request.args[arg] for arg in required_args}

    rs = requests.post(
        urljoin(args['server_url'], '/api/tts'),
        data=args['text'],
        params={
            'voice': args['voice'],
        },
    )

    return rs.content


# vim:sw=4:ts=4:et:



import json
import threading

from queue import Queue
from typing import Optional, List, Any, Dict, Union

from platypush.message.response import Response
from platypush.plugins.mqtt import MqttPlugin, action
from platypush.plugins.switch import SwitchPlugin


class ZigbeeMqttPlugin(MqttPlugin, SwitchPlugin):   # lgtm [py/missing-call-to-init]
    """
    This plugin allows you to interact with Zigbee devices over MQTT through any Zigbee sniffer and
    `zigbee2mqtt <https://www.zigbee2mqtt.io/>`_.

    In order to get started you'll need:

        - A Zigbee USB adapter/sniffer (in this example I'll use the
          `CC2531 <https://hackaday.io/project/163487-zigbee-cc2531-smart-home-usb-adapter>`_.
        - A Zigbee debugger/emulator + downloader cable (only to flash the firmware).

    Instructions:

        - Install `cc-tool <https://github.com/dashesy/cc-tool>`_ either from sources or from a package manager.
        - Connect the Zigbee to your PC/RaspberryPi in this way: ::

            USB -> CC debugger -> downloader cable -> CC2531 -> USB

        - The debugger and the adapter should be connected *at the same time*. If the later ``cc-tool`` command throws
          up an error, put the device in sync while connected by pressing the _Reset_ button on the debugger.
        - Check where the device is mapped. On Linux it will usually be ``/dev/ttyACM0``.
        - Download the latest `Z-Stack firmware <https://github.com/Koenkk/Z-Stack-firmware/tree/master/coordinator>`_
          to your device. Instructions for a CC2531 device:

          .. code-block:: shell

              wget https://github.com/Koenkk/Z-Stack-firmware/raw/master/coordinator/Z-Stack_Home_1.2/bin/default/CC2531_DEFAULT_20201127.zip
              unzip CC2531_DEFAULT_20201127.zip
              [sudo] cc-tool -e -w CC2531ZNP-Prod.hex

        - You can disconnect your debugger and downloader cable once the firmware is flashed.

        - Install ``zigbee2mqtt``. First install a node/npm environment, then either install ``zigbee2mqtt`` manually or
          through your package manager. **NOTE**: many API breaking changes have occurred on Zigbee2MQTT 1.17.0,
          therefore this integration will only be compatible with the version 1.17.0 of the service or higher versions.
          Manual instructions:

          .. code-block:: shell

              # Clone zigbee2mqtt repository
              [sudo] git clone https://github.com/Koenkk/zigbee2mqtt.git /opt/zigbee2mqtt
              [sudo] chown -R pi:pi /opt/zigbee2mqtt  # Or whichever is your user

              # Install dependencies (as user "pi")
              cd /opt/zigbee2mqtt
              npm install

        - You need to have an MQTT broker running somewhere. If not, you can install
            `Mosquitto <https://mosquitto.org/>`_ through your package manager on any device in your network.

        - Edit the ``/opt/zigbee2mqtt/data/configuration.yaml`` file to match the configuration of your MQTT broker:

          .. code-block:: yaml

              # MQTT settings
              mqtt:
                  # MQTT base topic for zigbee2mqtt MQTT messages
                  base_topic: zigbee2mqtt
                  # MQTT server URL
                  server: 'mqtt://localhost'
                  # MQTT server authentication, uncomment if required:
                  # user: my_user
                  # password: my_password

        - Also make sure that ``permit_join`` is set to ``True``, in order to allow Zigbee devices to join the network
          while you're configuring it. It's equally important to set ``permit_join`` to ``False`` once you have
          configured your network, to prevent accidental/malignant joins from outer Zigbee devices.

        - Start the ``zigbee2mqtt`` daemon on your device (the
          `official documentation <https://www.zigbee2mqtt.io/getting_started/running_zigbee2mqtt.html#5-optional-running-as-a-daemon-with-systemctl>`_
          also contains instructions on how to configure it as a ``systemd`` service:

          .. code-block:: shell

              cd /opt/zigbee2mqtt
              npm start

        - If you have Zigbee devices that are paired to other bridges, unlink them or do a factory reset to pair them
          to your new bridge.

        - If it all goes fine, once the daemon is running and a new device is found you should see traces like this in
          the output of ``zigbee2mqtt``::

            zigbee2mqtt:info  2019-11-09T12:19:56: Successfully interviewed '0x00158d0001dc126a', device has
            successfully been paired

        - You are now ready to use this integration.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)

    """

    def __init__(self, host: str = 'localhost', port: int = 1883, base_topic: str = 'zigbee2mqtt', timeout: int = 10,
                 tls_certfile: Optional[str] = None, tls_keyfile: Optional[str] = None,
                 tls_version: Optional[str] = None, tls_ciphers: Optional[str] = None,
                 username: Optional[str] = None, password: Optional[str] = None, **kwargs):
        """
        :param host: Default MQTT broker where ``zigbee2mqtt`` publishes its messages (default: ``localhost``).
        :param port: Broker listen port (default: 1883).
        :param base_topic: Topic prefix, as specified in ``/opt/zigbee2mqtt/data/configuration.yaml``
            (default: '``base_topic``').
        :param timeout: If the command expects from a response, then this timeout value will be used
            (default: 60 seconds).
        :param tls_cafile: If the connection requires TLS/SSL, specify the certificate authority file
            (default: None)
        :param tls_certfile: If the connection requires TLS/SSL, specify the certificate file (default: None)
        :param tls_keyfile: If the connection requires TLS/SSL, specify the key file (default: None)
        :param tls_version: If the connection requires TLS/SSL, specify the minimum TLS supported version
            (default: None)
        :param tls_ciphers: If the connection requires TLS/SSL, specify the supported ciphers (default: None)
        :param username: If the connection requires user authentication, specify the username (default: None)
        :param password: If the connection requires user authentication, specify the password (default: None)
        """
        super().__init__(host=host, port=port, tls_certfile=tls_certfile, tls_keyfile=tls_keyfile,
                         tls_version=tls_version, tls_ciphers=tls_ciphers, username=username,
                         password=password, **kwargs)

        self.base_topic = base_topic
        self.timeout = timeout
        self._info = {
            'devices': {},
            'groups': {},
        }

    def _get_network_info(self, **kwargs):
        self.logger.info('Fetching Zigbee network information')
        client = None
        mqtt_args = self._mqtt_args(**kwargs)
        timeout = 30
        if 'timeout' in mqtt_args:
            timeout = mqtt_args.pop('timeout')

        info = {
            'state': None,
            'info': {},
            'config': {},
            'devices': [],
            'groups': [],
        }

        info_ready_events = {topic: threading.Event() for topic in info.keys()}

        def _on_message():
            def callback(_, __, msg):
                topic = msg.topic.split('/')[-1]
                if topic in info:
                    info[topic] = msg.payload.decode() if topic == 'state' else json.loads(msg.payload.decode())
                    info_ready_events[topic].set()

            return callback

        try:
            host = mqtt_args.pop('host')
            port = mqtt_args.pop('port')
            client = self._get_client(**mqtt_args)
            client.on_message = _on_message()
            client.connect(host, port, keepalive=timeout)
            client.subscribe(self.base_topic + '/bridge/#')
            client.loop_start()

            for event in info_ready_events.values():
                info_ready = event.wait(timeout=timeout)
                if not info_ready:
                    raise TimeoutError('A timeout occurred while fetching the Zigbee network information')

            # Cache the new results
            self._info['devices'] = {
                device.get('friendly_name', device['ieee_address']): device
                for device in info.get('devices', [])
            }

            self._info['groups'] = {
                group.get('name'): group
                for group in info.get('groups', [])
            }

            self.logger.info('Zigbee network configuration updated')
            return info
        finally:
            try:
                client.loop_stop()
                client.disconnect()
            except Exception as e:
                self.logger.warning('Error on MQTT client disconnection: {}'.format(str(e)))

    def _topic(self, topic):
        return self.base_topic + '/' + topic

    @staticmethod
    def _parse_response(response: Union[dict, Response]) -> dict:
        if isinstance(response, Response):
            response = response.output

        assert response.get('status') != 'error', response.get('error', 'zigbee2mqtt error')
        return response

    @action
    def devices(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get the list of devices registered to the service.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return: List of paired devices. Example output:

        .. code-block:: json

            [
                {
                    "date_code": "20190608",
                    "friendly_name": "Coordinator",
                    "ieee_address": "0x00123456789abcde",
                    "network_address": 0,
                    "supported": false,
                    "type": "Coordinator",
                    "interviewing": false,
                    "interviewing_completed": true,
                    "definition": null,
                    "endpoints": {
                        "13": {
                            "bindings": [],
                            "clusters": {
                                "input": ["genOta"],
                                "output": []
                            },
                            "output": []
                        }
                    }
                },

                {
                    "date_code": "20180906",
                    "friendly_name": "My Lightbulb",
                    "ieee_address": "0x00123456789abcdf",
                    "network_address": 52715,
                    "power_source": "Mains (single phase)",
                    "software_build_id": "5.127.1.26581",
                    "model_id": "LCT001",
                    "supported": true,
                    "interviewing": false,
                    "interviewing_completed": true,
                    "type": "Router",
                    "definition": {
                        "description": "Hue white and color ambiance E26/E27/E14",
                        "model": "9290012573A",
                        "vendor": "Philips",
                        "exposes": [
                            {
                                "features": [
                                    {
                                        "access": 7,
                                        "description": "On/off state of this light",
                                        "name": "state",
                                        "property": "state",
                                        "type": "binary",
                                        "value_off": "OFF",
                                        "value_on": "ON",
                                        "value_toggle": "TOGGLE"
                                    },
                                    {
                                        "access": 7,
                                        "description": "Brightness of this light",
                                        "name": "brightness",
                                        "property": "brightness",
                                        "type": "numeric",
                                        "value_max": 254,
                                        "value_min": 0
                                    },
                                    {
                                        "access": 7,
                                        "description": "Color temperature of this light",
                                        "name": "color_temp",
                                        "property": "color_temp",
                                        "type": "numeric",
                                        "unit": "mired",
                                        "value_max": 500,
                                        "value_min": 150
                                    },
                                    {
                                        "description": "Color of this light in the CIE 1931 color space (x/y)",
                                        "features": [
                                            {
                                                "access": 7,
                                                "name": "x",
                                                "property": "x",
                                                "type": "numeric"
                                            },
                                            {
                                                "access": 7,
                                                "name": "y",
                                                "property": "y",
                                                "type": "numeric"
                                            }
                                        ],
                                        "name": "color_xy",
                                        "property": "color",
                                        "type": "composite"
                                    }
                                ],
                                "type": "light"
                            },
                            {
                                "access": 2,
                                "description": "Triggers an effect on the light (e.g. make light blink for a few seconds)",
                                "name": "effect",
                                "property": "effect",
                                "type": "enum",
                                "values": [
                                    "blink",
                                    "breathe",
                                    "okay",
                                    "channel_change",
                                    "finish_effect",
                                    "stop_effect"
                                ]
                            },
                            {
                                "access": 1,
                                "description": "Link quality (signal strength)",
                                "name": "linkquality",
                                "property": "linkquality",
                                "type": "numeric",
                                "unit": "lqi",
                                "value_max": 255,
                                "value_min": 0
                            }
                        ]
                    },

                    "endpoints": {
                        "11": {
                            "bindings": [],
                            "clusters": {
                                "input": [
                                    "genBasic",
                                    "genIdentify",
                                    "genGroups",
                                    "genScenes",
                                    "genOnOff",
                                    "genLevelCtrl",
                                    "touchlink",
                                    "lightingColorCtrl",
                                    "manuSpecificUbisysDimmerSetup"
                                ],
                                "output": [
                                    "genOta"
                                ]
                            },
                            "configured_reportings": []
                        },
                        "242": {
                            "bindings": [],
                            "clusters": {
                                "input": [
                                    "greenPower"
                                ],
                                "output": [
                                    "greenPower"
                                ]
                            },
                            "configured_reportings": []
                        }
                    }
                }
            ]

        """
        return self._get_network_info(**kwargs).get('devices')

    @action
    def permit_join(self, permit: bool = True, timeout: Optional[float] = None, **kwargs):
        """
        Enable/disable devices from joining the network. This is not persistent (will not be saved to
        ``configuration.yaml``).

        :param permit: Set to True to allow joins, False otherwise.
        :param timeout: Allow/disallow joins only for this amount of time.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if timeout:
            return self._parse_response(
                self.publish(topic=self._topic('bridge/request/permit_join'),
                             msg={'value': permit, 'time': timeout},
                             reply_topic=self._topic('bridge/response/permit_join'),
                             **self._mqtt_args(**kwargs)))

        return self.publish(topic=self._topic('bridge/request/permit_join'),
                            msg={'value': permit},
                            **self._mqtt_args(**kwargs))

    @action
    def factory_reset(self, **kwargs):
        """
        Perform a factory reset of a device connected to the network, following the procedure required by the particular
        device (for instance, Hue bulbs require the Zigbee adapter to be close to the device while a button on the back
        of the bulb is pressed).

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.publish(topic=self._topic('bridge/request/touchlink/factory_reset'), msg='', **self._mqtt_args(**kwargs))

    @action
    def log_level(self, level: str, **kwargs):
        """
        Change the log level at runtime. This change will not be persistent.

        :param level: Possible values: 'debug', 'info', 'warn', 'error'.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/config/log_level'), msg={'value': level},
                         reply_topic=self._topic('bridge/response/config/log_level'),
                         **self._mqtt_args(**kwargs)))

    @action
    def device_set_option(self, device: str, option: str, value: Any, **kwargs):
        """
        Change the options of a device. Options can only be changed, not added or deleted.

        :param device: Display name of the device.
        :param option: Option name.
        :param value: New value.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/options'),
                         reply_topic=self._topic('bridge/response/device/options'),
                         msg={
                             'id': device,
                             'options': {
                                 option: value,
                             }
                         }, **self._mqtt_args(**kwargs)))

    @action
    def device_remove(self, device: str, force: bool = False, **kwargs):
        """
        Remove a device from the network.

        :param device: Display name of the device.
        :param force: Force the remove also if the removal wasn't acknowledged by the device. Note: a forced remove
            only removes the entry from the internal database, but the device is likely to connect again when
            restarted unless it's factory reset (default: False).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/remove'),
                         msg={'id': device, 'force': force},
                         reply_topic=self._topic('bridge/response/device/remove'),
                         **self._mqtt_args(**kwargs)))

    @action
    def device_ban(self, device: str, **kwargs):
        """
        Ban a device from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/ban'),
                         reply_topic=self._topic('bridge/response/device/ban'),
                         msg={'id': device},
                         **self._mqtt_args(**kwargs)))

    @action
    def device_whitelist(self, device: str, **kwargs):
        """
        Whitelist a device on the network. Note: once at least a device is whitelisted, all the other non-whitelisted
        devices will be removed from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/whitelist'),
                         reply_topic=self._topic('bridge/response/device/whitelist'),
                         msg={'id': device},
                         **self._mqtt_args(**kwargs)))

    @action
    def device_rename(self, name: str, device: Optional[str] = None, **kwargs):
        """
        Rename a device on the network.

        :param name: New name.
        :param device: Current name of the device to rename. If no name is specified then the rename will
            affect the last device that joined the network.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if name == device:
            self.logger.info('Old and new name are the same: nothing to do')
            return

        # noinspection PyUnresolvedReferences
        devices = self.devices().output
        assert not [dev for dev in devices if dev.get('friendly_name') == name], \
            'A device named {} already exists on the network'.format(name)

        if device:
            req = {
                'from': device,
                'to': name,
            }
        else:
            req = {
                'last': True,
                'to': name,
            }

        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/rename'),
                         msg=req,
                         reply_topic=self._topic('bridge/response/device/rename'),
                         **self._mqtt_args(**kwargs)))

    @staticmethod
    def build_device_get_request(values: List[Dict[str, Any]]) -> dict:
        def extract_value(value: dict, root: dict):
            if not value.get('access', 1) & 0x1:
                # Property not readable
                return

            if 'features' not in value:
                if 'property' in value:
                    root[value['property']] = 0 if value['type'] == 'numeric' else ''
                return

            if 'property' in value:
                root[value['property']] = root.get(value['property'], {})
                root = root[value['property']]

            for feature in value['features']:
                extract_value(feature, root)

        ret = {}
        for value in values:
            extract_value(value, root=ret)

        return ret

    # noinspection PyShadowingBuiltins
    @action
    def device_get(self, device: str, property: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the properties of a device. The returned keys vary depending on the device. For example, a light bulb
        may have the "``state``" and "``brightness``" properties, while an environment sensor may have the
        "``temperature``" and "``humidity``" properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be retrieved (default: all).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: Key->value map of the device properties.
        """
        kwargs = self._mqtt_args(**kwargs)

        if property:
            properties = self.publish(topic=self._topic(device) + '/get/' + property, reply_topic=self._topic(device),
                                      msg={property: ''}, **kwargs).output

            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        if device not in self._info.get('devices', {}):
            # Refresh devices info
            self._get_network_info(**kwargs)

        assert self._info.get('devices', {}).get(device), 'No such device: ' + device
        exposes = (self._info.get('devices', {}).get(device, {}).get('definition', {}) or {}).get('exposes', [])
        if not exposes:
            return {}

        return self.publish(topic=self._topic(device) + '/get', reply_topic=self._topic(device),
                            msg=self.build_device_get_request(exposes), **kwargs)

    @action
    def devices_get(self, devices: Optional[List[str]] = None, **kwargs) -> Dict[str, dict]:
        """
        Get the properties of the devices connected to the network.

        :param devices: If set, then only the status of these devices (by friendly name) will be retrieved (default:
            retrieve all).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: Key->value map of the device properties:

            .. code-block:: json

                {
                    "Bulb": {
                        "state": "ON",
                        "brightness": 254
                    },
                    "Sensor": {
                        "temperature": 22.5
                    }
                }
        """
        kwargs = self._mqtt_args(**kwargs)

        if not devices:
            # noinspection PyUnresolvedReferences
            devices = set([
                device['friendly_name'] or device['ieee_address']
                for device in self.devices(**kwargs).output
            ])

        def worker(device: str, q: Queue):
            # noinspection PyUnresolvedReferences
            q.put(self.device_get(device, **kwargs).output)

        queues = {}
        workers = {}
        response = {}

        for device in devices:
            queues[device] = Queue()
            workers[device] = threading.Thread(target=worker, args=(device, queues[device]))
            workers[device].start()

        for device in devices:
            try:
                response[device] = queues[device].get(timeout=kwargs.get('timeout'))
                workers[device].join(timeout=kwargs.get('timeout'))
            except Exception as e:
                self.logger.warning('An error while getting the status of the device {}: {}'.format(
                    device, str(e)))

        return response

    @action
    def status(self, device: Optional[str] = None, *args, **kwargs):
        """
        Get the status of a device (by friendly name) or of all the connected devices (it wraps :meth:`.devices_get`).

        :param device: Device friendly name (default: get all devices).
        """
        return self.devices_get([device], *args, **kwargs)

    # noinspection PyShadowingBuiltins,DuplicatedCode
    @action
    def device_set(self, device: str, property: str, value: Any, **kwargs):
        """
        Set a properties on a device. The compatible properties vary depending on the device. For example, a light bulb
        may have the "``state``" and "``brightness``" properties, while an environment sensor may have the
        "``temperature``" and "``humidity``" properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        properties = self.publish(topic=self._topic(device + '/set'),
                                  reply_topic=self._topic(device),
                                  msg={property: value}, **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    @action
    def device_check_ota_updates(self, device: str, **kwargs) -> dict:
        """
        Check if the specified device has any OTA updates available to install.

        :param device: Address or friendly name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return:

            .. code-block:: json

                {
                    "id": "<device ID>",
                    "update_available": true,
                    "status": "ok"
                }

        """
        ret = self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/ota_update/check'),
                         reply_topic=self._topic('bridge/response/device/ota_update/check'),
                         msg={'id': device}, **self._mqtt_args(**kwargs)))

        return {
            'status': ret['status'],
            'id': ret.get('data', {}).get('id'),
            'update_available': ret.get('data', {}).get('update_available', False),
        }

    @action
    def device_install_ota_updates(self, device: str, **kwargs):
        """
        Install OTA updates for a device if available.

        :param device: Address or friendly name of the device.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/ota_update/update'),
                         reply_topic=self._topic('bridge/response/device/ota_update/update'),
                         msg={'id': device}, **self._mqtt_args(**kwargs)))

    @action
    def groups(self, **kwargs) -> List[dict]:
        """
        Get the groups registered on the device.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._get_network_info(**kwargs).get('groups')

    @action
    def info(self, **kwargs) -> dict:
        """
        Get the information, configuration and state of the network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return: Example:

            .. code-block:: json

                {
                    "state": "online",
                    "commit": "07cdc9d",
                    "config": {
                        "advanced": {
                            "adapter_concurrent": null,
                            "adapter_delay": null,
                            "availability_blacklist": [],
                            "availability_blocklist": [],
                            "availability_passlist": [],
                            "availability_timeout": 0,
                            "availability_whitelist": [],
                            "cache_state": true,
                            "cache_state_persistent": true,
                            "cache_state_send_on_startup": true,
                            "channel": 11,
                            "elapsed": false,
                            "ext_pan_id": [
                                221,
                                221,
                                221,
                                221,
                                221,
                                221,
                                221,
                                221
                            ],
                            "homeassistant_discovery_topic": "homeassistant",
                            "homeassistant_legacy_triggers": true,
                            "homeassistant_status_topic": "hass/status",
                            "last_seen": "disable",
                            "legacy_api": true,
                            "log_directory": "/opt/zigbee2mqtt/data/log/%TIMESTAMP%",
                            "log_file": "log.txt",
                            "log_level": "debug",
                            "log_output": [
                                "console",
                                "file"
                            ],
                            "log_rotation": true,
                            "log_syslog": {},
                            "pan_id": 6754,
                            "report": false,
                            "soft_reset_timeout": 0,
                            "timestamp_format": "YYYY-MM-DD HH:mm:ss"
                        },
                        "ban": [],
                        "blocklist": [],
                        "device_options": {},
                        "devices": {
                            "0x00123456789abcdf": {
                                "friendly_name": "My Lightbulb"
                            }
                        },
                        "experimental": {
                            "output": "json"
                        },
                        "external_converters": [],
                        "groups": {},
                        "homeassistant": false,
                        "map_options": {
                            "graphviz": {
                                "colors": {
                                    "fill": {
                                        "coordinator": "#e04e5d",
                                        "enddevice": "#fff8ce",
                                        "router": "#4ea3e0"
                                    },
                                    "font": {
                                        "coordinator": "#ffffff",
                                        "enddevice": "#000000",
                                        "router": "#ffffff"
                                    },
                                    "line": {
                                        "active": "#009900",
                                        "inactive": "#994444"
                                    }
                                }
                            }
                        },
                        "mqtt": {
                            "base_topic": "zigbee2mqtt",
                            "force_disable_retain": false,
                            "include_device_information": false,
                            "server": "mqtt://localhost"
                        },
                        "passlist": [],
                        "permit_join": true,
                        "serial": {
                            "disable_led": false,
                            "port": "/dev/ttyUSB0"
                        },
                        "whitelist": []
                    },
                    "coordinator": {
                        "meta": {
                            "maintrel": 3,
                            "majorrel": 2,
                            "minorrel": 6,
                            "product": 0,
                            "revision": 20190608,
                            "transportrev": 2
                        },
                        "type": "zStack12"
                    },
                    "log_level": "debug",
                    "network": {
                        "channel": 11,
                        "extended_pan_id": "0xdddddddddddddddd",
                        "pan_id": 6754
                    },
                    "permit_join": true,
                    "version": "1.17.0"
                }

        """
        info = self._get_network_info(**kwargs)
        return {
            'state': info.get('state'),
            'info': info.get('info'),
        }

    @action
    def group_add(self, name: str, id: Optional[int] = None, **kwargs):
        """
        Add a new group.

        :param name: Display name of the group.
        :param id: Optional numeric ID (default: auto-generated).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        payload = name if id is None else {
            'id': id,
            'friendly_name': name,
        }

        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/group/add'),
                         reply_topic=self._topic('bridge/response/group/add'),
                         msg=payload,
                         **self._mqtt_args(**kwargs))
        )

    @action
    def group_get(self, group: str, property: Optional[str] = None, **kwargs) -> dict:
        """
        Get one or more properties of a group. The compatible properties vary depending on the devices on the group.
        For example, a light bulb may have the "``state``" (with values ``"ON"`` and ``"OFF"``) and "``brightness``"
        properties, while an environment sensor may have the "``temperature``" and "``humidity``" properties, and so on.

        :param group: Display name of the group.
        :param property: Name of the property to retrieve (default: all available properties)
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        msg = {}
        if property:
            msg = {property: ''}

        properties = self.publish(topic=self._topic(group + '/get'),
                                  reply_topic=self._topic(group),
                                  msg=msg, **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    # noinspection PyShadowingBuiltins,DuplicatedCode
    @action
    def group_set(self, group: str, property: str, value: Any, **kwargs):
        """
        Set a properties on a group. The compatible properties vary depending on the devices on the group.
        For example, a light bulb may have the "``state``" (with values ``"ON"`` and ``"OFF"``) and "``brightness``"
        properties, while an environment sensor may have the "``temperature``" and "``humidity``" properties, and so on.

        :param group: Display name of the group.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        properties = self.publish(topic=self._topic(group + '/set'),
                                  reply_topic=self._topic(group),
                                  msg={property: value}, **self._mqtt_args(**kwargs)).output

        if property:
            assert property in properties, 'No such property: ' + property
            return {property: properties[property]}

        return properties

    @action
    def group_rename(self, name: str, group: str, **kwargs):
        """
        Rename a group.

        :param name: New name.
        :param group: Current name of the group to rename.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if name == group:
            self.logger.info('Old and new name are the same: nothing to do')
            return

        # noinspection PyUnresolvedReferences
        groups = {group.get('friendly_name'): group for group in self.groups().output}
        assert name not in groups, 'A group named {} already exists on the network'.format(name)

        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/group/rename'),
                         reply_topic=self._topic('bridge/response/group/rename'),
                         msg={'from': group, 'to': name} if group else name,
                         **self._mqtt_args(**kwargs)))

    @action
    def group_remove(self, name: str, **kwargs):
        """
        Remove a group.

        :param name: Display name of the group.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/group/remove'),
                         reply_topic=self._topic('bridge/response/group/remove'),
                         msg=name,
                         **self._mqtt_args(**kwargs)))

    @action
    def group_add_device(self, group: str, device: str, **kwargs):
        """
        Add a device to a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be added.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/group/members/add'),
                         reply_topic=self._topic('bridge/response/group/members/add'),
                         msg={
                             'group': group,
                             'device': device,
                         }, **self._mqtt_args(**kwargs)))

    @action
    def group_remove_device(self, group: str, device: Optional[str] = None, **kwargs):
        """
        Remove a device from a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be removed. If none is specified then all the devices registered
            to the specified group will be removed.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(
                topic=self._topic('bridge/request/group/members/remove{}'.format('_all' if device is None else '')),
                reply_topic=self._topic(
                    'bridge/response/group/members/remove{}'.format('_all' if device is None else '')),
                msg={
                    'group': group,
                    'device': device,
                }, **self._mqtt_args(**kwargs)))

    @action
    def bind_devices(self, source: str, target: str, **kwargs):
        """
        Bind two devices. Binding makes it possible that devices can directly control each other without the
        intervention of zigbee2mqtt or any home automation software. You may want to use this feature to bind
        for example an IKEA/Philips Hue dimmer switch to a light bulb, or a Zigbee remote to a thermostat.
        Read more on the `zigbee2mqtt binding page <https://www.zigbee2mqtt.io/information/binding.html>`_.

        :param source: Name of the source device. It can also be a group name, although the support is
            `still experimental <https://www.zigbee2mqtt.io/information/binding.html#binding-a-group>`_.
            You can also bind a specific device endpoint - for example ``MySensor/temperature``.
        :param target: Name of the target device.
            You can also bind a specific device endpoint - for example ``MyLight/state``.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/bind'),
                         reply_topic=self._topic('bridge/response/device/bind'),
                         msg={'from': source, 'to': target}, **self._mqtt_args(**kwargs)))

    @action
    def unbind_devices(self, source: str, target: str, **kwargs):
        """
        Un-bind two devices.

        :param source: Name of the source device.
            You can also bind a specific device endpoint - for example ``MySensor/temperature``.
        :param target: Name of the target device.
            You can also bind a specific device endpoint - for example ``MyLight/state``.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._parse_response(
            self.publish(topic=self._topic('bridge/request/device/unbind'),
                         reply_topic=self._topic('bridge/response/device/unbind'),
                         msg={'from': source, 'to': target}, **self._mqtt_args(**kwargs)))

    @action
    def on(self, device, *args, **kwargs) -> dict:
        """
        Implements :meth:`platypush.plugins.switch.plugin.SwitchPlugin.on` and turns on a Zigbee device with a writable
        binary property.
        """
        switch_info = self._get_switches_info().get(device)
        assert switch_info, '{} is not a valid switch'.format(device)
        props = self.device_set(device, switch_info['property'], switch_info['value_on']).output
        return self._properties_to_switch(device=device, props=props, switch_info=switch_info)

    @action
    def off(self, device, *args, **kwargs) -> dict:
        """
        Implements :meth:`platypush.plugins.switch.plugin.SwitchPlugin.off` and turns off a Zigbee device with a
        writable binary property.
        """
        switch_info = self._get_switches_info().get(device)
        assert switch_info, '{} is not a valid switch'.format(device)
        props = self.device_set(device, switch_info['property'], switch_info['value_off']).output
        return self._properties_to_switch(device=device, props=props, switch_info=switch_info)

    @action
    def toggle(self, device, *args, **kwargs) -> dict:
        """
        Implements :meth:`platypush.plugins.switch.plugin.SwitchPlugin.toggle` and toggles a Zigbee device with a
        writable binary property.
        """
        switch_info = self._get_switches_info().get(device)
        assert switch_info, '{} is not a valid switch'.format(device)
        props = self.device_set(device, switch_info['property'], switch_info['value_toggle']).output
        return self._properties_to_switch(device=device, props=props, switch_info=switch_info)

    @staticmethod
    def _properties_to_switch(device: str, props: dict, switch_info: dict) -> dict:
        return {
            'on': props[switch_info['property']] == switch_info['value_on'],
            'friendly_name': device,
            'name': device,
            **props,
        }

    def _get_switches_info(self) -> dict:
        def switch_info(device_info: dict) -> dict:
            exposes = (device_info.get('definition', {}) or {}).get('exposes', [])
            for exposed in exposes:
                for feature in exposed.get('features', []):
                    if feature.get('type') == 'binary' and 'value_on' in feature and 'value_off' in feature and \
                            feature.get('access', 0) & 2:
                        return {
                            'property': feature['property'],
                            'value_on': feature['value_on'],
                            'value_off': feature['value_off'],
                            'value_toggle': feature.get('value_toggle', None),
                        }

            return {}

        # noinspection PyUnresolvedReferences
        devices = self.devices().output
        switches_info = {}

        for device in devices:
            info = switch_info(device)
            if not info:
                continue

            switches_info[device.get('friendly_name', device.get('ieee_address'))] = info

        return switches_info

    @property
    def switches(self) -> List[dict]:
        """
        Implements the :class:`platypush.plugins.switch.SwitchPlugin.switches` property and returns the state of any
        device on the Zigbee network identified as a switch (a device is identified as a switch if it exposes a writable
        ``state`` property that can be set to ``ON`` or ``OFF``).
        """
        switches_info = self._get_switches_info()
        # noinspection PyUnresolvedReferences
        return [
            self._properties_to_switch(device=name, props=switch, switch_info=switches_info[name])
            for name, switch in self.devices_get(list(switches_info.keys())).output.items()
        ]


# vim:sw=4:ts=4:et:



import email
from typing import Optional, List, Dict, Union, Any, Tuple

from imapclient import IMAPClient
from imapclient.response_types import Address

from platypush.plugins import action
from platypush.plugins.mail import MailInPlugin, ServerInfo, Mail


class MailImapPlugin(MailInPlugin):
    """
    Plugin to interact with a mail server over IMAP.

    Requires:

        * **imapclient** (``pip install imapclient``)

    """

    _default_port = 143
    _default_ssl_port = 993

    def __init__(self, server: str, port: Optional[int] = None, username: Optional[str] = None,
                 password: Optional[str] = None, password_cmd: Optional[str] = None, access_token: Optional[str] = None,
                 oauth_mechanism: Optional[str] = 'XOAUTH2', oauth_vendor: Optional[str] = None, ssl: bool = False,
                 keyfile: Optional[str] = None, certfile: Optional[str] = None, timeout: Optional[int] = 60, **kwargs):
        """
        :param server: Server name/address.
        :param port: Port (default: 143 for plain, 993 for SSL).
        :param username: IMAP username.
        :param password: IMAP password.
        :param password_cmd: If you don't want to input your password in the configuration, run this command to fetch
            or decrypt the password.
        :param access_token: OAuth2 access token if the server supports OAuth authentication.
        :param oauth_mechanism: OAuth2 mechanism (default: ``XOAUTH2``).
        :param oauth_vendor: OAuth2 vendor (default: None).
        :param ssl: Use SSL (default: False).
        :param keyfile: Private key file for SSL connection if client authentication is required.
        :param certfile: SSL certificate file or chain.
        :param timeout: Server connect/read timeout in seconds (default: 60).
        """
        super().__init__(**kwargs)
        self.server_info = self._get_server_info(server=server, port=port, username=username, password=password,
                                                 password_cmd=password_cmd, ssl=ssl, keyfile=keyfile, certfile=certfile,
                                                 access_token=access_token, oauth_mechanism=oauth_mechanism,
                                                 oauth_vendor=oauth_vendor, timeout=timeout)

    def _get_server_info(self, server: Optional[str] = None, port: Optional[int] = None, username: Optional[str] = None,
                         password: Optional[str] = None, password_cmd: Optional[str] = None,
                         access_token: Optional[str] = None, oauth_mechanism: Optional[str] = None,
                         oauth_vendor: Optional[str] = None, ssl: Optional[bool] = None, keyfile: Optional[str] = None,
                         certfile: Optional[str] = None, timeout: Optional[int] = None, **kwargs) -> ServerInfo:
        return super()._get_server_info(server=server, port=port, username=username, password=password,
                                        password_cmd=password_cmd, ssl=ssl, keyfile=keyfile, certfile=certfile,
                                        default_port=self._default_port, default_ssl_port=self._default_ssl_port,
                                        access_token=access_token, oauth_mechanism=oauth_mechanism,
                                        oauth_vendor=oauth_vendor, timeout=timeout)

    def connect(self, **connect_args) -> IMAPClient:
        info = self._get_server_info(**connect_args)
        self.logger.info('Connecting to {}'.format(info.server))
        context = None

        if info.ssl:
            import ssl
            context = ssl.create_default_context()
            context.load_cert_chain(certfile=info.certfile, keyfile=info.keyfile)

        client = IMAPClient(host=info.server, port=info.port, ssl=info.ssl, ssl_context=context)
        if info.password:
            client.login(info.username, info.password)
        elif info.access_token:
            client.oauth2_login(info.username, access_token=info.access_token, mech=info.oauth_mechanism,
                                vendor=info.oauth_vendor)

        return client

    @staticmethod
    def _get_folders(data: List[tuple]) -> List[Dict[str, str]]:
        folders = []
        for line in data:
            (flags), delimiter, mailbox_name = line
            folders.append({
                'name': mailbox_name,
                'flags': [flag.decode() for flag in flags],
                'delimiter': delimiter.decode(),
            })

        return folders

    @action
    def get_folders(self, folder: str = '', pattern: str = '*', **connect_args) -> List[Dict[str, str]]:
        """
        Get the list of all the folders hosted on the server or those matching a pattern.

        :param folder: Base folder (default: root).
        :param pattern: Pattern to search (default: None).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        :return: Example:

        .. code-block:: json

            [
                {
                    "name": "INBOX",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                },
                {
                    "name": "Archive",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                },
                {
                    "name": "Spam",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                }
            ]

        """
        with self.connect(**connect_args) as client:
            data = client.list_folders(directory=folder, pattern=pattern)

        return self._get_folders(data)

    @action
    def get_sub_folders(self, folder: str = '', pattern: str = '*', **connect_args) -> List[Dict[str, str]]:
        """
        Get the list of all the sub-folders hosted on the server or those matching a pattern.

        :param folder: Base folder (default: root).
        :param pattern: Pattern to search (default: None).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        :return: Example:

        .. code-block:: json

            [
                {
                    "name": "INBOX",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                },
                {
                    "name": "Archive",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                },
                {
                    "name": "Spam",
                    "flags": "\\Noinferiors",
                    "delimiter": "/"
                }
            ]

        """
        with self.connect(**connect_args) as client:
            data = client.list_sub_folders(directory=folder, pattern=pattern)

        return self._get_folders(data)

    @staticmethod
    def _parse_address(imap_addr: Address) -> Dict[str, str]:
        return {
            'name': imap_addr.name.decode() if imap_addr.name else None,
            'route': imap_addr.route.decode() if imap_addr.route else None,
            'email': '{name}@{host}'.format(name=imap_addr.mailbox.decode(), host=imap_addr.host.decode())
        }

    @classmethod
    def _parse_addresses(cls, addresses: Optional[Tuple[Address]] = None) -> Dict[str, Dict[str, str]]:
        ret = {}
        if addresses:
            for addr in addresses:
                addr = cls._parse_address(addr)
                ret[addr['email']] = addr

        return ret

    @classmethod
    def _parse_message(cls, msg_id: int, imap_msg: Dict[bytes, Any]) -> Mail:
        message = {
            'id': msg_id,
            'seq': imap_msg[b'SEQ'],
        }

        if imap_msg.get(b'FLAGS'):
            message['flags'] = [flag.decode() for flag in imap_msg[b'FLAGS'] if flag]
        if b'INTERNALDATE' in imap_msg:
            message['internal_date'] = imap_msg[b'INTERNALDATE']
        if b'RFC822.SIZE' in imap_msg:
            message['size'] = imap_msg[b'RFC822.SIZE']
        if b'ENVELOPE' in imap_msg:
            envelope = imap_msg[b'ENVELOPE']
            message['bcc'] = cls._parse_addresses(envelope.bcc)
            message['cc'] = cls._parse_addresses(envelope.cc)
            message['date'] = envelope.date
            message['from'] = cls._parse_addresses(envelope.from_)
            message['message_id'] = envelope.message_id.decode() if envelope.message_id else None
            message['in_reply_to'] = envelope.in_reply_to.decode() if envelope.in_reply_to else None
            message['reply_to'] = cls._parse_addresses(envelope.reply_to)
            message['sender'] = cls._parse_addresses(envelope.sender)
            message['subject'] = envelope.subject.decode() if envelope.subject else None
            message['to'] = cls._parse_addresses(envelope.to)

        return Mail(**message)

    @action
    def search(self, criteria: Union[str, List[str]] = 'ALL', folder: str = 'INBOX',
               attributes: Optional[List[str]] = None, **connect_args) -> List[Mail]:
        """
        Search for messages on the server that fit the specified criteria.

        :param criteria: It should be a sequence of one or more criteria items. Each criteria item may be either unicode
            or bytes (default: ``ALL``). Example values::

                ['UNSEEN']
                ['SMALLER', 500]
                ['NOT', 'DELETED']
                ['TEXT', 'foo bar', 'FLAGGED', 'SUBJECT', 'baz']
                ['SINCE', '2020-03-14T12:13:45+00:00']

            It is also possible (but not recommended) to pass the combined criteria as a single string. In this case
            IMAPClient won't perform quoting, allowing lower-level specification of criteria. Examples of this style::

                'UNSEEN'
                'SMALLER 500'
                'NOT DELETED'
                'TEXT "foo bar" FLAGGED SUBJECT "baz"'
                'SINCE 03-Apr-2005'

            To support complex search expressions, criteria lists can be nested. The following will match messages that
            are both not flagged and do not have "foo" in the subject::

                ['NOT', ['SUBJECT', 'foo', 'FLAGGED']]

        :param folder: Folder to search (default: ``INBOX``).
        :param attributes: Attributes that should be retrieved, according to
            `RFC 3501 <https://tools.ietf.org/html/rfc3501>`_
            (default: ``ALL`` = ``[FLAGS INTERNALDATE RFC822.SIZE ENVELOPE]``).
            Note that ``BODY`` will be ignored if specified here for performance reasons - use :meth:`.get_message` if
            you want to get the full content of a message known its ID from :meth:`.search`.

        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        :return: List of messages matching the criteria. Example:

        .. code-block:: json

            [
              {
                "id": 702,
                "seq": 671,
                "flags": [
                  "nonjunk"
                ],
                "internal_date": "2020-08-30T00:31:52+00:00",
                "size": 2908738,
                "bcc": {},
                "cc": {},
                "date": "2020-08-30T00:31:52+00:00",
                "from": {
                  "test123@gmail.com": {
                    "name": "A test",
                    "route": null,
                    "email": "test123@gmail.com"
                  }
                },
                "message_id": "<SOMETHING@mail.gmail.com>",
                "in_reply_to": "<SOMETHING@mail.gmail.com>",
                "reply_to": {},
                "sender": {
                  "test123@gmail.com": {
                    "name": "A test",
                    "route": null,
                    "email": "test123@gmail.com"
                  }
                },
                "subject": "Test email",
                "to": {
                  "me@gmail.com": {
                    "name": null,
                    "route": null,
                    "email": "me@gmail.com"
                  }
                }
              }
            ]

        """
        if not attributes:
            attributes = ['ALL']
        else:
            attributes = [attr.upper() for attr in attributes]

        data = {}
        with self.connect(**connect_args) as client:
            client.select_folder(folder, readonly=True)
            ids = client.search(criteria)
            if len(ids):
                data = client.fetch(list(ids), attributes)

        return [
            self._parse_message(msg_id, data[msg_id])
            for msg_id in sorted(data.keys())
        ]

    @action
    def search_unseen_messages(self, folder: str = 'INBOX', **connect_args) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the unread messages.
        """
        return self.search(criteria='UNSEEN', directory=folder, attributes=['ALL'], **connect_args)

    @action
    def search_flagged_messages(self, folder: str = 'INBOX', **connect_args) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the flagged/starred messages.
        """
        return self.search(criteria='Flagged', directory=folder, attributes=['ALL'], **connect_args)

    @action
    def search_starred_messages(self, folder: str = 'INBOX', **connect_args) -> List[Mail]:
        """
        Shortcut for :meth:`.search` that returns only the flagged/starred messages.
        """
        return self.search_flagged_messages(folder, **connect_args)

    @action
    def sort(self, folder: str = 'INBOX', sort_criteria: Union[str, List[str]] = 'ARRIVAL',
             criteria: Union[str, List[str]] = 'ALL', **connect_args) -> List[int]:
        """
        Return a list of message ids from the currently selected folder, sorted by ``sort_criteria`` and optionally
        filtered by ``criteria``. Note that SORT is an extension to the IMAP4 standard so it may not be supported by
        all IMAP servers.

        :param folder: Folder to be searched (default: ``INBOX``).
        :param sort_criteria: It may be a sequence of strings or a single string. IMAPClient will take care any required
            conversions. Valid *sort_criteria* values::

              .. code-block:: python

                ['ARRIVAL']
                ['SUBJECT', 'ARRIVAL']
                'ARRIVAL'
                'REVERSE SIZE'

        :param criteria: Optional filter for the messages, as specified in :meth:`.search`.
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        :return: A list of message IDs that fit the criteria.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder, readonly=True)
            msg_ids = client.sort(sort_criteria=sort_criteria, criteria=criteria)
            return msg_ids

    @action
    def get_message(self, message: int, folder: str = 'INBOX', **connect_args) -> Mail:
        """
        Get the full content of a message given the ID returned by :meth:`.search`.

        :param message: Message ID.
        :param folder: Folder name (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        :return: A message in the same format as :meth:`.search`, with an added ``payload`` attribute containing the
            body/payload.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder, readonly=True)
            data = client.fetch(message, ['ALL', 'RFC822'])
            assert message in data, 'No such message ID: {}'.format(message)

            data = data[message]
            ret = self._parse_message(message, data)
            msg = email.message_from_bytes(data[b'RFC822'])
            ret.payload = msg.get_payload()

        return ret

    @action
    def create_folder(self, folder: str, **connect_args):
        """
        Create a folder on the server.

        :param folder: Folder name.
        :param connect_args:
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.create_folder(folder)

    @action
    def rename_folder(self, old_name: str, new_name: str, **connect_args):
        """
        Rename a folder on the server.

        :param old_name: Previous name
        :param new_name: New name
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.rename_folder(old_name, new_name)

    @action
    def delete_folder(self, folder: str, **connect_args):
        """
        Delete a folder from the server.

        :param folder: Folder name.
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.delete_folder(folder)

    @staticmethod
    def _convert_flags(flags: Union[str, List[str]]) -> List[bytes]:
        if isinstance(flags, str):
            flags = [flag.strip() for flag in flags.split(',')]
        return [('\\' + flag).encode() for flag in flags]

    @action
    def add_flags(self, messages: List[int], flags: Union[str, List[str]], folder: str = 'INBOX', **connect_args):
        """
        Add a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder)
            client.add_flags(messages, self._convert_flags(flags))

    @action
    def set_flags(self, messages: List[int], flags: Union[str, List[str]], folder: str = 'INBOX', **connect_args):
        """
        Set a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder)
            client.set_flags(messages, self._convert_flags(flags))

    @action
    def remove_flags(self, messages: List[int], flags: Union[str, List[str]], folder: str = 'INBOX', **connect_args):
        """
        Remove a set of flags to the specified set of message IDs.

        :param messages: List of message IDs.
        :param flags: List of flags to be added. Examples:

          .. code-block:: python

            ['Flagged']
            ['Seen', 'Deleted']
            ['Junk']

        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder)
            client.remove_flags(messages, self._convert_flags(flags))

    @action
    def flag_messages(self, messages: List[int], folder: str = 'INBOX', **connect_args):
        """
        Add a flag/star to the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        return self.add_flags(messages, ['Flagged'], folder=folder, **connect_args)

    @action
    def unflag_messages(self, messages: List[int], folder: str = 'INBOX', **connect_args):
        """
        Remove a flag/star from the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        return self.remove_flags(messages, ['Flagged'], folder=folder, **connect_args)

    @action
    def flag_message(self, message: int, folder: str = 'INBOX', **connect_args):
        """
        Add a flag/star to the specified set of message ID.

        :param message: Message ID.
        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        return self.flag_messages([message], folder=folder, **connect_args)

    @action
    def unflag_message(self, message: int, folder: str = 'INBOX', **connect_args):
        """
        Remove a flag/star from the specified set of message ID.

        :param message: Message ID.
        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        return self.unflag_messages([message], folder=folder, **connect_args)

    @action
    def delete_messages(self, messages: List[int], folder: str = 'INBOX', expunge: bool = True, **connect_args):
        """
        Set a specified set of message IDs as deleted.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param expunge: If set then the messages will also be expunged from the folder, otherwise they will only be
            marked as deleted (default: ``True``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        self.add_flags(messages, ['Deleted'], folder=folder, **connect_args)
        if expunge:
            self.expunge_messages(folder=folder, messages=messages, **connect_args)

    @action
    def undelete_messages(self, messages: List[int], folder: str = 'INBOX', **connect_args):
        """
        Remove the ``Deleted`` flag from the specified set of message IDs.

        :param messages: List of message IDs.
        :param folder: IMAP folder (default: ``INBOX``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        return self.remove_flags(messages, ['Deleted'], folder=folder, **connect_args)

    @action
    def copy_messages(self, messages: List[int], dest_folder: str, source_folder: str = 'INBOX', **connect_args):
        """
        Copy a set of messages IDs from a folder to another.

        :param messages: List of message IDs.
        :param source_folder: Source folder.
        :param dest_folder: Destination folder.
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(source_folder)
            client.copy(messages, dest_folder)

    @action
    def move_messages(self, messages: List[int], dest_folder: str, source_folder: str = 'INBOX', **connect_args):
        """
        Move a set of messages IDs from a folder to another.

        :param messages: List of message IDs.
        :param source_folder: Source folder.
        :param dest_folder: Destination folder.
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(source_folder)
            client.move(messages, dest_folder)

    @action
    def expunge_messages(self, folder: str = 'INBOX', messages: Optional[List[int]] = None, **connect_args):
        """
        When ``messages`` is not set, remove all the messages from ``folder`` marked as ``Deleted``.

        :param folder: IMAP folder (default: ``INBOX``).
        :param messages: List of message IDs to expunge (default: all those marked as ``Deleted``).
        :param connect_args: Arguments to pass to :meth:`._get_server_info` for server configuration override.
        """
        with self.connect(**connect_args) as client:
            client.select_folder(folder)
            client.expunge(messages)


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Optional

from platypush.message.event import Event


class HttpLogEvent(Event):
    """
    Event triggered when a new HTTP log entry is created.
    """
    def __init__(self, logfile: str, address: str, time: datetime, method: str, url: str, status: int, size: int,
                 http_version: str = '1.0', user_id: Optional[str] = None, user_identifier: Optional[str] = None,
                 referrer: Optional[str] = None, user_agent: Optional[str] = None, **kwargs):
        super().__init__(logfile=logfile, address=address, time=time, method=method, url=url, status=status, size=size,
                         http_version=http_version, user_id=user_id, user_identifier=user_identifier, referrer=referrer,
                         user_agent=user_agent, **kwargs)



import array
import struct
import time
from fcntl import ioctl
from typing import IO

from platypush.backend import Backend
from platypush.message.event.joystick import JoystickConnectedEvent, JoystickDisconnectedEvent, \
    JoystickButtonPressedEvent, JoystickButtonReleasedEvent, JoystickAxisEvent


class JoystickLinuxBackend(Backend):
    """
    This backend intercepts events from joystick devices through the native Linux API implementation.

    It is loosely based on https://gist.github.com/rdb/8864666, which itself uses the
    `Linux kernel joystick API <https://www.kernel.org/doc/Documentation/input/joystick-api.txt>`_ to interact with
    the devices.

    Triggers:

        * :class:`platypush.message.event.joystick.JoystickConnectedEvent` when the joystick is connected.
        * :class:`platypush.message.event.joystick.JoystickDisconnectedEvent` when the joystick is disconnected.
        * :class:`platypush.message.event.joystick.JoystickButtonPressedEvent` when a joystick button is pressed.
        * :class:`platypush.message.event.joystick.JoystickButtonReleasedEvent` when a joystick button is released.
        * :class:`platypush.message.event.joystick.JoystickAxisEvent` when an axis value of the joystick changes.

    """

    # These constants were borrowed from linux/input.h
    axis_names = {
        0x00: 'x',
        0x01: 'y',
        0x02: 'z',
        0x03: 'rx',
        0x04: 'ry',
        0x05: 'rz',
        0x06: 'throttle',
        0x07: 'rudder',
        0x08: 'wheel',
        0x09: 'gas',
        0x0a: 'brake',
        0x10: 'hat0x',
        0x11: 'hat0y',
        0x12: 'hat1x',
        0x13: 'hat1y',
        0x14: 'hat2x',
        0x15: 'hat2y',
        0x16: 'hat3x',
        0x17: 'hat3y',
        0x18: 'pressure',
        0x19: 'distance',
        0x1a: 'tilt_x',
        0x1b: 'tilt_y',
        0x1c: 'tool_width',
        0x20: 'volume',
        0x28: 'misc',
    }

    button_names = {
        0x120: 'trigger',
        0x121: 'thumb',
        0x122: 'thumb2',
        0x123: 'top',
        0x124: 'top2',
        0x125: 'pinkie',
        0x126: 'base',
        0x127: 'base2',
        0x128: 'base3',
        0x129: 'base4',
        0x12a: 'base5',
        0x12b: 'base6',
        0x12f: 'dead',
        0x130: 'a',
        0x131: 'b',
        0x132: 'c',
        0x133: 'x',
        0x134: 'y',
        0x135: 'z',
        0x136: 'tl',
        0x137: 'tr',
        0x138: 'tl2',
        0x139: 'tr2',
        0x13a: 'select',
        0x13b: 'start',
        0x13c: 'mode',
        0x13d: 'thumbl',
        0x13e: 'thumbr',
        0x220: 'dpad_up',
        0x221: 'dpad_down',
        0x222: 'dpad_left',
        0x223: 'dpad_right',
        # XBox 360 controller uses these codes.
        0x2c0: 'dpad_left',
        0x2c1: 'dpad_right',
        0x2c2: 'dpad_up',
        0x2c3: 'dpad_down',
    }

    def __init__(self, device: str = '/dev/input/js0', *args, **kwargs):
        """
        :param device: Joystick device to monitor (default: ``/dev/input/js0``).
        """
        super().__init__(*args, **kwargs)
        self.device = device
        self._axis_states = {}
        self._button_states = {}
        self._axis_map = []
        self._button_map = []

    def _init_joystick(self, dev: IO):
        # Get the device name.
        buf = array.array('B', [0] * 64)
        ioctl(dev, 0x80006a13 + (0x10000 * len(buf)), buf)  # JSIOCGNAME(len)
        js_name = buf.tobytes().rstrip(b'\x00').decode('utf-8')

        # Get number of axes and buttons.
        buf = array.array('B', [0])
        ioctl(dev, 0x80016a11, buf)  # JSIOCGAXES
        num_axes = buf[0]

        buf = array.array('B', [0])
        ioctl(dev, 0x80016a12, buf)  # JSIOCGBUTTONS
        num_buttons = buf[0]

        # Get the axis map.
        buf = array.array('B', [0] * 0x40)
        ioctl(dev, 0x80406a32, buf)  # JSIOCGAXMAP

        for axis in buf[:num_axes]:
            axis_name = self.axis_names.get(axis, 'unknown(0x%02x)' % axis)
            self._axis_map.append(axis_name)
            self._axis_states[axis_name] = 0.0

        # Get the button map.
        buf = array.array('H', [0] * 200)
        ioctl(dev, 0x80406a34, buf)  # JSIOCGBTNMAP

        for btn in buf[:num_buttons]:
            btn_name = self.button_names.get(btn, 'unknown(0x%03x)' % btn)
            self._button_map.append(btn_name)
            self._button_states[btn_name] = 0

        self.bus.post(JoystickConnectedEvent(device=self.device, name=js_name, axes=self._axis_map,
                                             buttons=self._button_map))

    def run(self):
        super().run()
        self.logger.info(f'Opening {self.device}...')

        while not self.should_stop():
            # Open the joystick device.
            try:
                jsdev = open(self.device, 'rb')
                self._init_joystick(jsdev)
            except Exception as e:
                self.logger.debug(f'Joystick device on {self.device} not available: {e}')
                time.sleep(5)
                continue

            # Joystick event loop
            while not self.should_stop():
                try:
                    evbuf = jsdev.read(8)
                    if evbuf:
                        _, value, evt_type, number = struct.unpack('IhBB', evbuf)

                        if evt_type & 0x80:  # Initial state notification
                            continue

                        if evt_type & 0x01:
                            button = self._button_map[number]
                            if button:
                                self._button_states[button] = value
                                evt_class = JoystickButtonPressedEvent if value else JoystickButtonReleasedEvent
                                # noinspection PyTypeChecker
                                self.bus.post(evt_class(device=self.device, button=button))

                        if evt_type & 0x02:
                            axis = self._axis_map[number]
                            if axis:
                                fvalue = value / 32767.0
                                self._axis_states[axis] = fvalue
                                # noinspection PyTypeChecker
                                self.bus.post(JoystickAxisEvent(device=self.device, axis=axis, value=fvalue))
                except OSError as e:
                    self.logger.warning(f'Connection to {self.device} lost: {e}')
                    self.bus.post(JoystickDisconnectedEvent(device=self.device))
                    break



import os

from marshmallow import fields
from marshmallow.schema import Schema

from platypush.schemas import StrippedString


class IRCServerSchema(Schema):
    server = fields.String(
        required=True,
        metadata=dict(
            description='Server address or hostname',
            example='irc.example.org',
        )
    )

    port = fields.Int(
        missing=6667,
        metadata=dict(
            description='IRC server port',
            example=6667,
        )
    )

    password = fields.String(allow_none=True, default=None, metadata=dict(example='password'))
    nickname = fields.String(required=True, metadata=dict(example='testbot'))
    realname = fields.String(allow_none=True, metadata=dict(example='My Real Name'))
    alias = StrippedString(
        metadata=dict(description='Friendly name for this bot/server connection')
    )

    channels = fields.List(
        fields.String(),
        missing=list,
        metadata=dict(
            description='List of channels the bot will connect to',
            example=['#channel1', '#channel2', '#channel3']
        )
    )

    ssl = fields.Boolean(missing=False)
    ipv6 = fields.Boolean(missing=False)
    stop_message = StrippedString(missing='Application stopped', metadata=dict(description='Quit/die message'))

    dcc_ip_whitelist = fields.List(
        fields.String,
        missing=list,
        metadata=dict(
            description='If specified then only DCC connections from the IP addresses on this list will be accepted',
        )
    )

    dcc_ip_blacklist = fields.List(
        fields.String,
        missing=list,
        metadata=dict(
            description='If specified then DCC connections from the IP addresses on this list will be rejected',
        )
    )

    dcc_nick_whitelist = fields.List(
        fields.String,
        missing=list,
        metadata=dict(
            description='If specified then only DCC connections from the nicknames on this list will be accepted',
        )
    )

    dcc_nick_blacklist = fields.List(
        fields.String,
        missing=list,
        metadata=dict(
            description='If specified then DCC connections from the nicknames on this list will be rejected',
        )
    )

    dcc_downloads_dir = fields.String(
        missing=os.path.join(os.path.expanduser('~'), 'Downloads'),
        metadata=dict(description='DCC file transfers will be downloaded to this folder (default: ~/Downloads)'),
    )

    response_timeout = fields.Number(
        missing=30.,
        metadata=dict(
            description='How long we should wait for a response to an IRC request '
                        '(default: 30 seconds)',
        )
    )

    dcc_file_transfer_timeout = fields.Number(
        missing=30.,
        metadata=dict(
            description='How long we should wait on a pending DCC file transfer with '
                        'no data being transmitted (default: 30 seconds)',
        )
    )

    dcc_accept_timeout = fields.Number(
        missing=300.,
        metadata=dict(
            description='How long we should wait on a pending DCC request '
                        'until the user accepts (default: 300 seconds)',
        )
    )

    dcc_max_connections = fields.Int(
        missing=10,
        metadata=dict(
            description='Maximum number of concurrent DCC connections allowed on this bot (default: 10)'
        )
    )


class IRCServerStatusSchema(Schema):
    server = StrippedString(required=True)
    port = fields.Int(required=True)
    alias = StrippedString()
    real_name = fields.String()
    nickname = fields.String()
    is_connected = fields.Boolean()
    connected_channels = fields.List(fields.String)


class IRCChannelSchema(Schema):
    name = fields.String(required=True)
    modes = fields.List(fields.String)
    opers = fields.List(fields.String)
    owners = fields.List(fields.String)
    users = fields.List(fields.String)
    voiced = fields.List(fields.String)
    is_invite_only = fields.Boolean()
    is_moderated = fields.Boolean()
    is_protected = fields.Boolean()
    is_secret = fields.Boolean()
